@@ -1,4 +1,4 @@
-/* This file is (c) 2008-2011 Konstantin Isakov <ikm@users.berlios.de>
+/* This file is (c) 2008-2012 Konstantin Isakov <ikm@goldendict.org>
  * Part of GoldenDict. Licensed under GPLv3 or later, see the LICENSE file */
 
 #include "btreeidx.hh"
@@ -10,6 +10,14 @@
 #include <math.h>
 #include <string.h>
 #include <stdlib.h>
+#include "gddebug.hh"
+#include "wstring_qt.hh"
+#include "qt4x5.hh"
+
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+#include <QRegularExpression>
+#include "wildcard.hh"
+#endif
 
 //#define __BTREE_USE_LZO
 // LZO mode is experimental and unsupported. Tests didn't show any substantial
@@ -36,11 +44,12 @@ namespace BtreeIndexing {
 
 using gd::wstring;
 using gd::wchar;
+using std::pair;
 
 enum
 {
   BtreeMinElements = 64,
-  BtreeMaxElements = 2048
+  BtreeMaxElements = 8192
 };
 
 BtreeIndex::BtreeIndex():
@@ -74,35 +83,48 @@ void BtreeIndex::openIndex( IndexInfo const & indexInfo,
   rootNode.clear();
 }
 
-vector< WordArticleLink > BtreeIndex::findArticles( wstring const & str )
+vector< WordArticleLink > BtreeIndex::findArticles( wstring const & word, bool ignoreDiacritics )
 {
   vector< WordArticleLink > result;
 
-  wstring folded = Folding::apply( str );
-
-  bool exactMatch;
-
-  vector< char > leaf;
-  uint32_t nextLeaf;
-
-  char const * leafEnd;
-
-  char const * chainOffset = findChainOffsetExactOrPrefix( folded, exactMatch,
-                                                           leaf, nextLeaf,
-                                                           leafEnd );
-
-  if ( chainOffset && exactMatch )
+  try
   {
-    result = readChain( chainOffset );
+    wstring folded = Folding::apply( word );
+    if( folded.empty() )
+      folded = Folding::applyWhitespaceOnly( word );
 
-    antialias( str, result );
+    bool exactMatch;
+
+    vector< char > leaf;
+    uint32_t nextLeaf;
+
+    char const * leafEnd;
+
+    char const * chainOffset = findChainOffsetExactOrPrefix( folded, exactMatch,
+                                                             leaf, nextLeaf,
+                                                             leafEnd );
+
+    if ( chainOffset && exactMatch )
+    {
+      result = readChain( chainOffset );
+
+      antialias( word, result, ignoreDiacritics );
+    }
+  }
+  catch( std::exception & e )
+  {
+    gdWarning( "Articles searching failed, error: %s\n", e.what() );
+    result.clear();
+  }
+  catch(...)
+  {
+    qWarning( "Articles searching failed\n" );
+    result.clear();
   }
 
   return result;
 }
 
-class BtreeWordSearchRequest;
-  
 class BtreeWordSearchRunnable: public QRunnable
 {
   BtreeWordSearchRequest & r;
@@ -123,59 +145,34 @@ public:
   virtual void run();
 };
 
-class BtreeWordSearchRequest: public Dictionary::WordSearchRequest
-{
-  friend class BtreeWordSearchRunnable;
-
-  BtreeDictionary & dict;
-  wstring str;
-  unsigned long maxResults;
-  unsigned minLength;
-  int maxSuffixVariation;
-  bool allowMiddleMatches;
-  QAtomicInt isCancelled;
-  QSemaphore hasExited;
-
-public:
-
-  BtreeWordSearchRequest( BtreeDictionary & dict_,
-                          wstring const & str_,
-                          unsigned minLength_,
-                          int maxSuffixVariation_,
-                          bool allowMiddleMatches_,
-                          unsigned long maxResults_ ):
-    dict( dict_ ), str( str_ ),
-    maxResults( maxResults_ ),
-    minLength( minLength_ ),
-    maxSuffixVariation( maxSuffixVariation_ ),
-    allowMiddleMatches( allowMiddleMatches_ )
-  {
-    QThreadPool::globalInstance()->start(
-      new BtreeWordSearchRunnable( *this, hasExited ) );
-  }
-
-  void run(); // Run from another thread by BtreeWordSearchRunnable
-
-  virtual void cancel()
-  {
-    isCancelled.ref();
-  }
-  
-  ~BtreeWordSearchRequest()
-  {
-    isCancelled.ref();
-    hasExited.acquire();
-  }
-};
-
 void BtreeWordSearchRunnable::run()
 {
   r.run();
 }
 
-void BtreeWordSearchRequest::run()
+BtreeWordSearchRequest::BtreeWordSearchRequest( BtreeDictionary & dict_,
+                                                wstring const & str_,
+                                                unsigned minLength_,
+                                                int maxSuffixVariation_,
+                                                bool allowMiddleMatches_,
+                                                unsigned long maxResults_,
+                                                bool startRunnable ):
+  dict( dict_ ), str( str_ ),
+  maxResults( maxResults_ ),
+  minLength( minLength_ ),
+  maxSuffixVariation( maxSuffixVariation_ ),
+  allowMiddleMatches( allowMiddleMatches_ )
 {
-  if ( isCancelled )
+  if( startRunnable )
+  {
+    QThreadPool::globalInstance()->start(
+      new BtreeWordSearchRunnable( *this, hasExited ) );
+  }
+}
+
+void BtreeWordSearchRequest::findMatches()
+{
+  if ( Qt4x5::AtomicInt::loadAcquire( isCancelled ) )
   {
     finish();
     return;
@@ -188,8 +185,123 @@ void BtreeWordSearchRequest::run()
     return;
   }
   
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+  QRegularExpression regexp;
+#else
+  QRegExp regexp;
+#endif
+  bool useWildcards = false;
+  if( allowMiddleMatches )
+    useWildcards = ( str.find( '*' ) != wstring::npos ||
+                     str.find( '?' ) != wstring::npos ||
+                     str.find( '[' ) != wstring::npos ||
+                     str.find( ']' ) != wstring::npos );
+
   wstring folded = Folding::apply( str );
 
+  int minMatchLength = 0;
+
+  if( useWildcards )
+  {
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+    regexp.setPattern( wildcardsToRegexp( gd::toQString( Folding::applyDiacriticsOnly( Folding::applySimpleCaseOnly( str ) ) ) ) );
+    if( !regexp.isValid() )
+      regexp.setPattern( QRegularExpression::escape( regexp.pattern() ) );
+    regexp.setPatternOptions( QRegularExpression::CaseInsensitiveOption );
+#else
+    regexp.setPattern( gd::toQString( Folding::applyDiacriticsOnly( Folding::applySimpleCaseOnly( str ) ) ) );
+    regexp.setPatternSyntax( QRegExp::WildcardUnix );
+    regexp.setCaseSensitivity( Qt::CaseInsensitive );
+#endif
+
+    bool bNoLetters = folded.empty();
+    wstring foldedWithWildcards;
+
+    if( bNoLetters )
+      foldedWithWildcards = Folding::applyWhitespaceOnly( str );
+    else
+      foldedWithWildcards = Folding::apply( str, useWildcards );
+
+    // Calculate minimum match length
+
+    bool insideSet = false;
+    bool escaped = false;
+    for( wstring::size_type x = 0; x < foldedWithWildcards.size(); x++ )
+    {
+      wchar ch = foldedWithWildcards[ x ];
+
+      if( ch == L'\\' && !escaped )
+      {
+        escaped = true;
+        continue;
+      }
+
+      if( ch == L']' && !escaped )
+      {
+        insideSet = false;
+        continue;
+      }
+
+      if( insideSet )
+      {
+        escaped = false;
+        continue;
+      }
+
+      if( ch == L'[' && !escaped )
+      {
+        minMatchLength += 1;
+        insideSet = true;
+        continue;
+      }
+
+      if( ch == L'*' && !escaped )
+        continue;
+
+      escaped = false;
+      minMatchLength += 1;
+    }
+
+    // Fill first match chars
+
+    folded.clear();
+    folded.reserve( foldedWithWildcards.size() );
+    escaped = false;
+    for( wstring::size_type x = 0; x < foldedWithWildcards.size(); x++ )
+    {
+      wchar ch = foldedWithWildcards[ x ];
+
+      if( escaped )
+      {
+        if( bNoLetters || ( ch != L'*' && ch != L'?' && ch != L'[' && ch != L']' ) )
+          folded.push_back( ch );
+        escaped = false;
+        continue;
+      }
+
+      if( ch == L'\\' )
+      {
+        if( bNoLetters || folded.empty() )
+        {
+          escaped = true;
+          continue;
+        }
+        else
+          break;
+      }
+
+      if( ch == '*' || ch == '?' || ch == '[' || ch == ']' )
+        break;
+
+      folded.push_back( ch );
+    }
+  }
+  else
+  {
+    if( folded.empty() )
+      folded = Folding::applyWhitespaceOnly( str );
+  }
+
   int initialFoldedSize = folded.size();
 
   int charsLeftToChop = 0;
@@ -205,105 +317,172 @@ void BtreeWordSearchRequest::run()
       charsLeftToChop = maxSuffixVariation;
   }
 
-  for( ; ; )
+  try
   {
-    bool exactMatch;
-  
-    vector< char > leaf;
-    uint32_t nextLeaf;
-    char const * leafEnd;
-  
-    char const * chainOffset = dict.findChainOffsetExactOrPrefix( folded, exactMatch,
-                                                                  leaf, nextLeaf,
-                                                                  leafEnd );
-  
-    if ( chainOffset )
     for( ; ; )
     {
-      if ( isCancelled )
-        break;
-      
-      //printf( "offset = %u, size = %u\n", chainOffset - &leaf.front(), leaf.size() );
-  
-      vector< WordArticleLink > chain = dict.readChain( chainOffset );
-  
-      wstring chainHead = Utf8::decode( chain[ 0 ].word );
-  
-      wstring resultFolded = Folding::apply( chainHead );
-  
-      if ( resultFolded.size() >= folded.size() && !resultFolded.compare( 0, folded.size(), folded ) )
-      {
-        // Exact or prefix match
+      bool exactMatch;
+      vector< char > leaf;
+      uint32_t nextLeaf;
+      char const * leafEnd;
 
-        Mutex::Lock _( dataMutex );
-  
-        for( unsigned x = 0; x < chain.size(); ++x )
-        {
-          // Skip middle matches, if requested. If suffix variation is specified,
-          // make sure the string isn't larger than requested.
-          if ( ( allowMiddleMatches || Folding::apply( Utf8::decode( chain[ x ].prefix ) ).empty() ) &&
-               ( maxSuffixVariation < 0 || (int)resultFolded.size() - initialFoldedSize <= maxSuffixVariation ) )
-              matches.push_back( Utf8::decode( chain[ x ].prefix + chain[ x ].word ) );
-        }
-  
-        if ( matches.size() >= maxResults )
-        {
-          // For now we actually allow more than maxResults if the last
-          // chain yield more than one result. That's ok and maybe even more
-          // desirable.
-          break;
-        }
-      }
-      else
-        // Neither exact nor a prefix match, end this
-        break;
-  
-      // Fetch new leaf if we're out of chains here
-  
-      if ( chainOffset >= leafEnd )
+      char const * chainOffset = dict.findChainOffsetExactOrPrefix( folded, exactMatch,
+                                                                    leaf, nextLeaf,
+                                                                    leafEnd );
+
+      if ( chainOffset )
+      for( ; ; )
       {
-        // We're past the current leaf, fetch the next one
-  
-        //printf( "advancing\n" );
-  
-        if ( nextLeaf )
+        if ( Qt4x5::AtomicInt::loadAcquire( isCancelled ) )
+          break;
+
+        //DPRINTF( "offset = %u, size = %u\n", chainOffset - &leaf.front(), leaf.size() );
+
+        vector< WordArticleLink > chain = dict.readChain( chainOffset );
+
+        wstring chainHead = Utf8::decode( chain[ 0 ].word );
+
+        wstring resultFolded = Folding::apply( chainHead );
+        if( resultFolded.empty() )
+          resultFolded = Folding::applyWhitespaceOnly( chainHead );
+
+        if ( ( useWildcards && folded.empty() ) ||
+             ( resultFolded.size() >= folded.size()
+               && !resultFolded.compare( 0, folded.size(), folded ) ) )
         {
-          Mutex::Lock _( *dict.idxFileMutex );
-          
-          dict.readNode( nextLeaf, leaf );
-          leafEnd = &leaf.front() + leaf.size();
-  
-          nextLeaf = dict.idxFile->read< uint32_t >();
-          chainOffset = &leaf.front() + sizeof( uint32_t );
-  
-          uint32_t leafEntries = *(uint32_t *)&leaf.front();
-  
-          if ( leafEntries == 0xffffFFFF )
+          // Exact or prefix match
+
+          Mutex::Lock _( dataMutex );
+
+          for( unsigned x = 0; x < chain.size(); ++x )
           {
-            //printf( "bah!\n" );
-            exit( 1 );
+            if( useWildcards )
+            {
+              wstring word = Utf8::decode( chain[ x ].prefix + chain[ x ].word );
+              wstring result = Folding::applyDiacriticsOnly( word );
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+              if( result.size() >= (wstring::size_type)minMatchLength )
+              {
+                QRegularExpressionMatch match = regexp.match( gd::toQString( result ) );
+                if( match.hasMatch() && match.capturedStart() == 0 )
+                {
+                  addMatch( word );
+                }
+              }
+#else
+              if( result.size() >= (wstring::size_type)minMatchLength
+                  && regexp.indexIn( gd::toQString( result ) ) == 0
+                  && regexp.matchedLength() >= minMatchLength )
+              {
+                addMatch( word );
+              }
+#endif
+            }
+            else
+            {
+              // Skip middle matches, if requested. If suffix variation is specified,
+              // make sure the string isn't larger than requested.
+              if ( ( allowMiddleMatches || Folding::apply( Utf8::decode( chain[ x ].prefix ) ).empty() ) &&
+                   ( maxSuffixVariation < 0 || (int)resultFolded.size() - initialFoldedSize <= maxSuffixVariation ) )
+                  addMatch( Utf8::decode( chain[ x ].prefix + chain[ x ].word ) );
+            }
+          }
+
+          if( Qt4x5::AtomicInt::loadAcquire( isCancelled ) )
+            break;
+
+          if ( matches.size() >= maxResults )
+          {
+            // For now we actually allow more than maxResults if the last
+            // chain yield more than one result. That's ok and maybe even more
+            // desirable.
+            break;
           }
         }
         else
-          break; // That was the last leaf
-      }
-    }
+          // Neither exact nor a prefix match, end this
+          break;
 
-    if ( charsLeftToChop && !isCancelled )
-    {
-      --charsLeftToChop;
-      folded.resize( folded.size() - 1 );
+        // Fetch new leaf if we're out of chains here
+
+        if ( chainOffset >= leafEnd )
+        {
+          // We're past the current leaf, fetch the next one
+
+          //DPRINTF( "advancing\n" );
+
+          if ( nextLeaf )
+          {
+            Mutex::Lock _( *dict.idxFileMutex );
+
+            dict.readNode( nextLeaf, leaf );
+            leafEnd = &leaf.front() + leaf.size();
+
+            nextLeaf = dict.idxFile->read< uint32_t >();
+            chainOffset = &leaf.front() + sizeof( uint32_t );
+
+            uint32_t leafEntries = *(uint32_t *)&leaf.front();
+
+            if ( leafEntries == 0xffffFFFF )
+            {
+              //DPRINTF( "bah!\n" );
+              exit( 1 );
+            }
+          }
+          else
+            break; // That was the last leaf
+        }
+      }
+
+      if ( charsLeftToChop && !Qt4x5::AtomicInt::loadAcquire( isCancelled ) )
+      {
+        --charsLeftToChop;
+        folded.resize( folded.size() - 1 );
+      }
+      else
+        break;
     }
-    else
-      break;
   }
+  catch( std::exception & e )
+  {
+    qWarning( "Index searching failed: \"%s\", error: %s\n",
+              dict.getName().c_str(), e.what() );
+  }
+  catch(...)
+  {
+    gdWarning( "Index searching failed: \"%s\"\n", dict.getName().c_str() );
+  }
+}
+
+void BtreeWordSearchRequest::run()
+{
+  if ( Qt4x5::AtomicInt::loadAcquire( isCancelled ) )
+  {
+    finish();
+    return;
+  }
+
+  if ( dict.ensureInitDone().size() )
+  {
+    setErrorString( QString::fromUtf8( dict.ensureInitDone().c_str() ) );
+    finish();
+    return;
+  }
+
+  findMatches();
 
   finish();
 }
 
+BtreeWordSearchRequest::~BtreeWordSearchRequest()
+{
+  isCancelled.ref();
+  hasExited.acquire();
+}
+
 sptr< Dictionary::WordSearchRequest > BtreeDictionary::prefixMatch(
   wstring const & str, unsigned long maxResults )
-  throw( std::exception )
+  THROW_SPEC( std::exception )
 {
   return new BtreeWordSearchRequest( *this, str, 0, -1, true, maxResults );
 }
@@ -311,7 +490,7 @@ sptr< Dictionary::WordSearchRequest > BtreeDictionary::prefixMatch(
 sptr< Dictionary::WordSearchRequest > BtreeDictionary::stemmedMatch(
   wstring const & str, unsigned minLength, unsigned maxSuffixVariation,
   unsigned long maxResults )
-  throw( std::exception )
+  THROW_SPEC( std::exception )
 {
   return new BtreeWordSearchRequest( *this, str, minLength, (int)maxSuffixVariation,
                                      false, maxResults );
@@ -324,7 +503,7 @@ void BtreeIndex::readNode( uint32_t offset, vector< char > & out )
   uint32_t uncompressedSize = idxFile->read< uint32_t >();
   uint32_t compressedSize = idxFile->read< uint32_t >();
 
-  //printf( "%x,%x\n", uncompressedSize, compressedSize );
+  //DPRINTF( "%x,%x\n", uncompressedSize, compressedSize );
 
   out.resize( uncompressedSize );
 
@@ -367,9 +546,7 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
   
   // Lookup the index by traversing the index btree
 
-  vector< char > charBuffer;
   vector< wchar > wcharBuffer;
-  vector< char > wordsBuffer;
 
   exactMatch = false;
 
@@ -387,6 +564,38 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
   char const * leaf = &rootNode.front();
   leafEnd = leaf + rootNode.size();
 
+  if( target.empty() )
+  {
+    //For empty target string we return first chain in index
+    for( ; ; )
+    {
+      uint32_t leafEntries = *(uint32_t *)leaf;
+
+      if ( leafEntries == 0xffffFFFF )
+      {
+        // A node
+        currentNodeOffset = *( (uint32_t *)leaf + 1 );
+        readNode( currentNodeOffset, extLeaf );
+        leaf = &extLeaf.front();
+        leafEnd = leaf + extLeaf.size();
+        nextLeaf = idxFile->read< uint32_t >();
+      }
+      else
+      {
+        // A leaf
+        if( currentNodeOffset == rootOffset )
+        {
+          // Only one leaf in index, there's no next leaf
+          nextLeaf = 0;
+        }
+        if( !leafEntries )
+          return 0;
+
+        return leaf + sizeof( uint32_t );
+      }
+    }
+  }
+
   for( ; ; )
   {
     // Is it a leaf or a node?
@@ -397,7 +606,7 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
     {
       // A node
 
-      //printf( "=>a node\n" );
+      //DPRINTF( "=>a node\n" );
 
       uint32_t const * offsets = (uint32_t *)leaf + 1;
 
@@ -431,13 +640,13 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
           wcharBuffer.resize( wordSize + 1 );
   
         long result = Utf8::decode( closestString, wordSize, &wcharBuffer.front() );
-  
+
         if ( result < 0 )
           throw Utf8::exCantDecode( closestString );
   
         wcharBuffer[ result ] = 0;
 
-        //printf( "Checking against %s\n", closestString );
+        //DPRINTF( "Checking against %s\n", closestString );
 
         compareResult = target.compare( &wcharBuffer.front() );
   
@@ -468,7 +677,7 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
       }
 
       #if 0
-      printf( "The winner is %s, compareResult = %d\n", closestString, compareResult );
+      DPRINTF( "The winner is %s, compareResult = %d\n", closestString, compareResult );
 
       if ( closestString != ptr )
       {
@@ -477,19 +686,19 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
         while( left != ptr && left[ -1 ] )
           --left;
 
-        printf( "To the left: %s\n", left );
+        DPRINTF( "To the left: %s\n", left );
       }
       else
-        printf( "To the lest -- nothing\n" );
+        DPRINTF( "To the lest -- nothing\n" );
 
       char const * right = closestString + strlen( closestString ) + 1;
 
       if ( right != leafEnd )
       {
-        printf( "To the right: %s\n", right );
+        DPRINTF( "To the right: %s\n", right );
       }
       else
-        printf( "To the right -- nothing\n" );
+        DPRINTF( "To the right -- nothing\n" );
       #endif
 
       // Now, whatever the outcome (compareResult) is, we need to find
@@ -521,14 +730,14 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
         currentNodeOffset = offsets[ entry + 1 ];
       }
 
-      //printf( "reading node at %x\n", currentNodeOffset );
+      //DPRINTF( "reading node at %x\n", currentNodeOffset );
       readNode( currentNodeOffset, extLeaf );
       leaf = &extLeaf.front();
       leafEnd = leaf + extLeaf.size();
     }
     else
     {
-      //printf( "=>a leaf\n" );
+      //DPRINTF( "=>a leaf\n" );
       // A leaf
 
       // If this leaf is the root, there's no next leaf, it just can't be.
@@ -561,7 +770,7 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
 
           memcpy( &chainSize, ptr, sizeof( uint32_t ) );
 
-          //printf( "%s + %s\n", ptr + sizeof( uint32_t ), ptr + sizeof( uint32_t ) + strlen( ptr + sizeof( uint32_t ) ) + 1 );
+          //DPRINTF( "%s + %s\n", ptr + sizeof( uint32_t ), ptr + sizeof( uint32_t ) + strlen( ptr + sizeof( uint32_t ) ) + 1 );
 
           ptr += sizeof( uint32_t ) + chainSize;
         }
@@ -575,7 +784,7 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
 
       for( ; ; )
       {
-        //printf( "window = %u, ws = %u\n", window - &chainOffsets.front(), windowSize );
+        //DPRINTF( "window = %u, ws = %u\n", window - &chainOffsets.front(), windowSize );
 
         char const ** chainToCheck = window + windowSize/2;
         ptr = *chainToCheck;
@@ -588,7 +797,7 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
         if ( wcharBuffer.size() <= wordSize )
           wcharBuffer.resize( wordSize + 1 );
   
-        //printf( "checking agaist word %s, left = %u\n", ptr, leafEntries );
+        //DPRINTF( "checking against word %s, left = %u\n", ptr, leafEntries );
   
         long result = Utf8::decode( ptr, wordSize, &wcharBuffer.front() );
   
@@ -598,6 +807,8 @@ char const * BtreeIndex::findChainOffsetExactOrPrefix( wstring const & target,
         wcharBuffer[ result ] = 0;
   
         wstring foldedWord = Folding::apply( &wcharBuffer.front() );
+        if( foldedWord.empty() )
+          foldedWord = Folding::applyWhitespaceOnly( &wcharBuffer.front() );
   
         int compareResult = target.compare( foldedWord );
   
@@ -673,8 +884,6 @@ vector< WordArticleLink > BtreeIndex::readChain( char const * & ptr )
 
   vector< WordArticleLink > result;
 
-  vector< char > charBuffer;
-
   while( chainSize )
   {
     string str = ptr;
@@ -701,16 +910,22 @@ vector< WordArticleLink > BtreeIndex::readChain( char const * & ptr )
 }
 
 void BtreeIndex::antialias( wstring const & str,
-                            vector< WordArticleLink > & chain )
+                            vector< WordArticleLink > & chain,
+                            bool ignoreDiacritics )
 {
-  wstring caseFolded = Folding::applySimpleCaseOnly( str );
+  wstring caseFolded = Folding::applySimpleCaseOnly( gd::normalize( str ) );
+  if( ignoreDiacritics )
+    caseFolded = Folding::applyDiacriticsOnly( caseFolded );
 
   for( unsigned x = chain.size(); x--; )
   {
     // If after applying case folding to each word they wouldn't match, we
     // drop the entry.
-    if ( Folding::applySimpleCaseOnly( Utf8::decode( chain[ x ].prefix + chain[ x ].word ) ) !=
-         caseFolded )
+    wstring entry = Folding::applySimpleCaseOnly( gd::normalize( Utf8::decode( chain[ x ].prefix + chain[ x ].word ) ) );
+    if( ignoreDiacritics )
+      entry = Folding::applyDiacriticsOnly( entry );
+
+    if ( entry != caseFolded )
       chain.erase( chain.begin() + x );
     else
     if ( chain[ x ].prefix.size() ) // If there's a prefix, merge it with the word,
@@ -844,7 +1059,7 @@ static uint32_t buildBtreeNode( IndexedWords::const_iterator & nextIndex,
                          &compressedData.front(), &compressedSize, workMem )
        != LZO_E_OK )
   {
-    fprintf( stderr, "Failed to compress btree node.\n" );
+    FDPRINTF( stderr, "Failed to compress btree node.\n" );
     abort();
   }
 
@@ -857,7 +1072,7 @@ static uint32_t buildBtreeNode( IndexedWords::const_iterator & nextIndex,
   if ( compress( &compressedData.front(), &compressedSize,
                  &uncompressedData.front(), uncompressedData.size() ) != Z_OK )
   {
-    fprintf( stderr, "Failed to compress btree node.\n" );
+    qFatal( "Failed to compress btree node." );
     abort();
   }
 
@@ -893,11 +1108,31 @@ static uint32_t buildBtreeNode( IndexedWords::const_iterator & nextIndex,
   return offset;
 }
 
-void IndexedWords::addWord( wstring const & word, uint32_t articleOffset )
+void IndexedWords::addWord( wstring const & word, uint32_t articleOffset, unsigned int maxHeadwordSize )
 {
   wchar const * wordBegin = word.c_str();
   string::size_type wordSize = word.size();
 
+  // Safeguard us against various bugs here. Don't attempt adding words
+  // which are freakishly huge.
+  if ( wordSize > maxHeadwordSize )
+  {
+#define MAX_LOG_WORD_SIZE 500
+    string headword;
+    if( wordSize <= MAX_LOG_WORD_SIZE )
+      headword = Utf8::encode( word );
+    else
+    {
+      std::vector< char > buffer( MAX_LOG_WORD_SIZE * 4 );
+      headword = string( &buffer.front(),
+                         Utf8::encode( wordBegin, MAX_LOG_WORD_SIZE, &buffer.front() ) );
+      headword += "...";
+    }
+    gdWarning( "Skipped too long headword: \"%s\"", headword.c_str() );
+    return;
+#undef MAX_LOG_WORD_SIZE
+  }
+
   // Skip any leading whitespace
   while( *wordBegin && Folding::isWhitespace( *wordBegin ) )
   {
@@ -913,13 +1148,37 @@ void IndexedWords::addWord( wstring const & word, uint32_t articleOffset )
 
   vector< char > utfBuffer( wordSize * 4 );
 
+  int wordsAdded = 0; // Number of stored parts
+
   for( ; ; )
   {
     // Skip any whitespace/punctuation
     for( ; ; ++nextChar )
     {
-      if ( !*nextChar )
-        return; // End of string ends everything
+      if ( !*nextChar ) // End of string ends everything
+      {
+          if( wordsAdded == 0)
+          {
+              wstring folded = Folding::applyWhitespaceOnly( wstring( wordBegin, wordSize ) );
+              if( !folded.empty() )
+              {
+                  iterator i = insert(
+                    IndexedWords::value_type(
+                      string( &utfBuffer.front(),
+                              Utf8::encode( folded.data(), folded.size(), &utfBuffer.front() ) ),
+                      vector< WordArticleLink >() ) ).first;
+
+                  // Try to conserve memory somewhat -- slow insertions are ok
+                  i->second.reserve( i->second.size() + 1 );
+
+                  string utfWord( &utfBuffer.front(),
+                                  Utf8::encode( wordBegin, wordSize, &utfBuffer.front() ) );
+                  string utfPrefix;
+                  i->second.push_back( WordArticleLink( utfWord, articleOffset, utfPrefix ) );
+              }
+          }
+          return;
+      }
   
       if ( !Folding::isWhitespace( *nextChar ) && !Folding::isPunct( *nextChar ) )
         break;
@@ -948,6 +1207,8 @@ void IndexedWords::addWord( wstring const & word, uint32_t articleOffset )
       i->second.push_back( WordArticleLink( utfWord, articleOffset, utfPrefix ) );
     }
 
+    wordsAdded += 1;
+
     // Skip all non-whitespace/punctuation
     for( ++nextChar; ; ++nextChar )
     {
@@ -962,12 +1223,11 @@ void IndexedWords::addWord( wstring const & word, uint32_t articleOffset )
 
 void IndexedWords::addSingleWord( wstring const & word, uint32_t articleOffset )
 {
-  vector< WordArticleLink > links( 1, WordArticleLink( Utf8::encode( word ),
-                                                       articleOffset ) );
-
-  insert(
-      IndexedWords::value_type(
-          Utf8::encode( Folding::apply( word ) ), links ) );
+  wstring folded = Folding::apply( word );
+  if( folded.empty() )
+      folded = Folding::applyWhitespaceOnly( word );
+  operator []( Utf8::encode( folded ) ).push_back(
+    WordArticleLink( Utf8::encode( word ), articleOffset ) );
 }
 
 IndexInfo buildIndex( IndexedWords const & indexedWords, File::Class & file )
@@ -995,7 +1255,7 @@ IndexInfo buildIndex( IndexedWords const & indexedWords, File::Class & file )
   if ( btreeMaxElements > BtreeMaxElements )
     btreeMaxElements = BtreeMaxElements;
 
-  printf( "Building a tree of %u elements\n", btreeMaxElements );
+  GD_DPRINTF( "Building a tree of %u elements\n", (unsigned) btreeMaxElements );
 
 
   uint32_t lastLeafOffset = 0;
@@ -1007,4 +1267,300 @@ IndexInfo buildIndex( IndexedWords const & indexedWords, File::Class & file )
   return IndexInfo( btreeMaxElements, rootOffset );
 }
 
+void BtreeIndex::getAllHeadwords( QSet< QString > & headwords )
+{
+  if ( !idxFile )
+    throw exIndexWasNotOpened();
+
+  findArticleLinks( NULL, NULL, &headwords );
+}
+
+void BtreeIndex::findAllArticleLinks( QVector< WordArticleLink > & articleLinks )
+{
+  if ( !idxFile )
+    throw exIndexWasNotOpened();
+
+  QSet< uint32_t > offsets;
+
+  findArticleLinks( &articleLinks, &offsets, NULL );
+}
+
+void BtreeIndex::findArticleLinks( QVector< WordArticleLink > * articleLinks,
+                                   QSet< uint32_t > * offsets,
+                                   QSet< QString > *headwords,
+                                   QAtomicInt * isCancelled )
+{
+  uint32_t currentNodeOffset = rootOffset;
+  uint32_t nextLeaf = 0;
+  uint32_t leafEntries;
+
+  Mutex::Lock _( *idxFileMutex );
+
+  if ( !rootNodeLoaded )
+  {
+    // Time to load our root node. We do it only once, at the first request.
+    readNode( rootOffset, rootNode );
+    rootNodeLoaded = true;
+  }
+
+  char const * leaf = &rootNode.front();
+  char const * leafEnd = leaf + rootNode.size();
+  char const * chainPtr = 0;
+
+  vector< char > extLeaf;
+
+  // Find first leaf
+
+  for( ; ; )
+  {
+    leafEntries = *(uint32_t *)leaf;
+
+    if( isCancelled && Qt4x5::AtomicInt::loadAcquire( *isCancelled ) )
+      return;
+
+    if ( leafEntries == 0xffffFFFF )
+    {
+      // A node
+      currentNodeOffset = *( (uint32_t *)leaf + 1 );
+      readNode( currentNodeOffset, extLeaf );
+      leaf = &extLeaf.front();
+      leafEnd = leaf + extLeaf.size();
+      nextLeaf = idxFile->read< uint32_t >();
+    }
+    else
+    {
+      // A leaf
+      chainPtr = leaf + sizeof( uint32_t );
+      break;
+    }
+  }
+
+  if ( !leafEntries )
+  {
+    // Empty leaf? This may only be possible for entirely empty trees only.
+    if ( currentNodeOffset != rootOffset )
+      throw exCorruptedChainData();
+    else
+      return; // No match
+  }
+
+  // Read all chains
+
+  for( ; ; )
+  {
+    vector< WordArticleLink > result = readChain( chainPtr );
+
+    if( headwords && static_cast< vector< WordArticleLink >::size_type >( headwords->capacity() ) < headwords->size() + result.size() )
+    {
+      int n = headwords->capacity();
+      headwords->reserve( n + n / 10 );
+    }
+
+    if( offsets && static_cast< vector< WordArticleLink >::size_type >( offsets->capacity() ) < offsets->size() + result.size() )
+    {
+      int n = offsets->capacity();
+      offsets->reserve( n + n / 10 );
+    }
+
+    if( articleLinks && static_cast< vector< WordArticleLink >::size_type >( articleLinks->capacity() ) < articleLinks->size() + result.size() )
+    {
+      int n = articleLinks->capacity();
+      articleLinks->reserve( n + n / 10 );
+    }
+
+    for( unsigned i = 0; i < result.size(); i++ )
+    {
+      if( isCancelled && Qt4x5::AtomicInt::loadAcquire( *isCancelled ) )
+        return;
+
+      if( headwords )
+        headwords->insert( QString::fromUtf8( ( result[ i ].prefix + result[ i ].word ).c_str() ) );
+
+      if( offsets && offsets->contains( result[ i ].articleOffset ) )
+        continue;
+
+      if( offsets )
+        offsets->insert( result[ i ].articleOffset );
+
+      if( articleLinks )
+        articleLinks->push_back( WordArticleLink( result[ i ].prefix + result[ i ].word, result[ i ].articleOffset ) );
+    }
+
+    if ( chainPtr >= leafEnd )
+    {
+      // We're past the current leaf, fetch the next one
+
+      if ( nextLeaf )
+      {
+        readNode( nextLeaf, extLeaf );
+        leaf = &extLeaf.front();
+        leafEnd = leaf + extLeaf.size();
+
+        nextLeaf = idxFile->read< uint32_t >();
+        chainPtr = leaf + sizeof( uint32_t );
+
+        leafEntries = *(uint32_t *)leaf;
+
+        if ( leafEntries == 0xffffFFFF )
+          throw exCorruptedChainData();
+      }
+      else
+        break; // That was the last leaf
+    }
+  }
+}
+
+void BtreeIndex::getHeadwordsFromOffsets( QList<uint32_t> & offsets,
+                                          QVector<QString> & headwords,
+                                          QAtomicInt * isCancelled )
+{
+  uint32_t currentNodeOffset = rootOffset;
+  uint32_t nextLeaf = 0;
+  uint32_t leafEntries;
+
+  std::sort( offsets.begin(), offsets.end() );
+
+  Mutex::Lock _( *idxFileMutex );
+
+  if ( !rootNodeLoaded )
+  {
+    // Time to load our root node. We do it only once, at the first request.
+    readNode( rootOffset, rootNode );
+    rootNodeLoaded = true;
+  }
+
+  char const * leaf = &rootNode.front();
+  char const * leafEnd = leaf + rootNode.size();
+  char const * chainPtr = 0;
+
+  vector< char > extLeaf;
+
+  // Find first leaf
+
+  for( ; ; )
+  {
+    leafEntries = *(uint32_t *)leaf;
+
+    if( isCancelled && Qt4x5::AtomicInt::loadAcquire( *isCancelled ) )
+      return;
+
+    if ( leafEntries == 0xffffFFFF )
+    {
+      // A node
+      currentNodeOffset = *( (uint32_t *)leaf + 1 );
+      readNode( currentNodeOffset, extLeaf );
+      leaf = &extLeaf.front();
+      leafEnd = leaf + extLeaf.size();
+      nextLeaf = idxFile->read< uint32_t >();
+    }
+    else
+    {
+      // A leaf
+      chainPtr = leaf + sizeof( uint32_t );
+      break;
+    }
+  }
+
+  if ( !leafEntries )
+  {
+    // Empty leaf? This may only be possible for entirely empty trees only.
+    if ( currentNodeOffset != rootOffset )
+      throw exCorruptedChainData();
+    else
+      return; // No match
+  }
+
+  // Read all chains
+
+  typedef QList< uint32_t >::Iterator ListIterator;
+  ListIterator begOffsets = offsets.begin();
+  ListIterator endOffsets = offsets.end();
+
+  for( ; ; )
+  {
+    vector< WordArticleLink > result = readChain( chainPtr );
+
+    for( unsigned i = 0; i < result.size(); i++ )
+    {
+      std::pair< ListIterator, ListIterator > const range = std::equal_range( begOffsets, endOffsets,
+                                                                              result.at( i ).articleOffset );
+
+      if( range.first != range.second )
+      {
+        if( isCancelled && Qt4x5::AtomicInt::loadAcquire( *isCancelled ) )
+          return;
+
+        headwords.append(  QString::fromUtf8( ( result[ i ].prefix + result[ i ].word ).c_str() ) );
+        offsets.erase( range.first );
+        begOffsets = offsets.begin();
+        endOffsets = offsets.end();
+      }
+
+      if( offsets.isEmpty() )
+        break;
+    }
+
+    if( offsets.isEmpty() )
+      break;
+
+    if ( chainPtr >= leafEnd )
+    {
+      // We're past the current leaf, fetch the next one
+
+      if ( nextLeaf )
+      {
+        readNode( nextLeaf, extLeaf );
+        leaf = &extLeaf.front();
+        leafEnd = leaf + extLeaf.size();
+
+        nextLeaf = idxFile->read< uint32_t >();
+        chainPtr = leaf + sizeof( uint32_t );
+
+        leafEntries = *(uint32_t *)leaf;
+
+        if ( leafEntries == 0xffffFFFF )
+          throw exCorruptedChainData();
+      }
+      else
+        break; // That was the last leaf
+    }
+  }
+}
+
+bool BtreeDictionary::getHeadwords( QStringList &headwords )
+{
+  QSet< QString > setOfHeadwords;
+
+  headwords.clear();
+  setOfHeadwords.reserve( getWordCount() );
+
+  try
+  {
+    getAllHeadwords( setOfHeadwords );
+
+    if( setOfHeadwords.size() )
+    {
+#if QT_VERSION >= 0x040700
+      headwords.reserve( setOfHeadwords.size() );
+#endif
+
+      QSet< QString >::const_iterator it = setOfHeadwords.constBegin();
+      QSet< QString >::const_iterator end = setOfHeadwords.constEnd();
+
+      for( ; it != end; ++it )
+        headwords.append( *it );
+    }
+  }
+  catch( std::exception &ex )
+  {
+    gdWarning( "Failed headwords retrieving for \"%s\", reason: %s\n", getName().c_str(), ex.what() );
+  }
+
+  return headwords.size() > 0;
+}
+
+void BtreeDictionary::getArticleText(uint32_t, QString &, QString & )
+{
+}
+
 }
