@@ -1,11 +1,19 @@
-/* This file is (c) 2008-2011 Konstantin Isakov <ikm@users.berlios.de>
+/* This file is (c) 2008-2012 Konstantin Isakov <ikm@goldendict.org>
  * Part of GoldenDict. Licensed under GPLv3 or later, see the LICENSE file */
 
 #include "dsl_details.hh"
+
 #include "folding.hh"
 #include "langcoder.hh"
-#include <wctype.h>
+#include "gddebug.hh"
+#include "ufile.hh"
+#include "wstring_qt.hh"
+#include "utf8.hh"
+
 #include <stdio.h>
+#include <wctype.h>
+
+#include <algorithm>
 
 namespace Dsl {
 namespace Details {
@@ -13,7 +21,7 @@ namespace Details {
 using gd::wstring;
 using std::list;
 
-#ifdef __WIN32
+#ifndef __linux__
 
 // wcscasecmp() function is a GNU extension, we need to reimplement it
 // for non-GNU systems.
@@ -34,9 +42,115 @@ int wcscasecmp( const wchar *s1, const wchar *s2 )
 
 #endif
 
+static DSLLangCode LangCodes[] =
+{
+  { 1, "en" },
+  { 1033, "en" },
+  { 2, "ru" },
+  { 1049, "ru" },
+  { 1068, "az" },
+  { 1025, "ar" },
+  { 1067, "am" },
+  { 15, "af" },
+  { 1078, "af" },
+  { 9, "eu" },
+  { 1069, "eu" },
+  { 1133, "ba" },
+  { 21, "be" },
+  { 1059, "be" },
+  { 22, "bg" },
+  { 1026, "bg" },
+  { 19, "hu" },
+  { 1038, "hu" },
+  { 10, "nl" },
+  { 1043, "nl" },
+  { 1032, "el" },
+  { 1079, "ka" },
+  { 13, "da" },
+  { 1030, "da" },
+  { 16, "id" },
+  { 1057, "id" },
+  { 1039, "is" },
+  { 6, "es" },
+  { 7, "es" },
+  { 3082, "es" },
+  { 1034, "es" },
+  { 5, "it" },
+  { 1040, "it" },
+  { 1087, "kk" },
+  { 1595, "ky" },
+  { 28, "ch" },
+  { 29, "ch" },
+  { 1028, "ch" },
+  { 2052, "ch" },
+  { 30, "la" },
+  { 1540, "la" },
+  { 1142, "la" },
+  { 1062, "lv" },
+  { 1063, "lt" },
+  { 1086, "ms" },
+  { 3, "de" },
+  { 26, "de" },
+  { 1031, "de" },
+  { 32775, "de" },
+  { 14, "nb" },
+  { 1044, "nb" },
+  { 25, "nn" },
+  { 2068, "nn" },
+  { 20, "pl" },
+  { 1045, "pl" },
+  { 8, "pt" },
+  { 2070, "pt" },
+  { 1048, "ro" },
+  { 23, "sr" },
+  { 3098, "sr" },
+  { 1051, "sk" },
+  { 1060, "sl" },
+  { 17, "sw" },
+  { 1089, "sw" },
+  { 1064, "tg" },
+  { 1092, "tt" },
+  { 27, "tr" },
+  { 1055, "tr" },
+  { 1090, "tk" },
+  { 1091, "tz" },
+  { 24, "uk" },
+  { 1058, "uk" },
+  { 11, "fi" },
+  { 1035, "fi" },
+  { 4, "fr" },
+  { 1036, "fr" },
+  { 18, "cs" },
+  { 1029, "cs" },
+  { 12, "sv" },
+  { 1053, "sv" },
+  { 1061, "et" },
+  { 0, "" },
+};
+
+string findCodeForDslId( int id )
+{
+  for( DSLLangCode const * lc = LangCodes; lc->code_id; ++lc )
+  {
+    if ( id == lc->code_id )
+    {
+      // We've got a match
+      return string( lc->code );
+    }
+  }
+  return string();
+}
+
+bool isAtSignFirst( wstring const & str )
+{
+  // Test if '@' is first in string except spaces and dsl tags
+  QRegExp reg( "[ \\t]*(?:\\[[^\\]]+\\][ \\t]*)*@", Qt::CaseInsensitive, QRegExp::RegExp2 );
+  return reg.indexIn( gd::toQString( str ) ) == 0;
+}
+
 /////////////// ArticleDom
 
-wstring ArticleDom::Node::renderAsText() const
+wstring ArticleDom::Node::renderAsText( bool stripTrsTag ) const
 {
   if ( !isTag )
     return text;
@@ -44,21 +158,51 @@ wstring ArticleDom::Node::renderAsText() const
   wstring result;
 
   for( list< Node >::const_iterator i = begin(); i != end(); ++i )
-    result += i->renderAsText();
+    if( !stripTrsTag || i->tagName != GD_NATIVE_TO_WS( L"!trs" ) )
+      result += i->renderAsText( stripTrsTag );
 
   return result;
 }
 
-// Returns true if src == 'm' and dest is 'mX', where X is a digit
-static inline bool checkM( wstring const & dest, wstring const & src )
+namespace {
+
+/// @return true if @p tagName equals "mN" where N is a digit
+bool is_mN( wstring const & tagName )
 {
-  return ( src == GD_NATIVE_TO_WS( L"m" ) && dest.size() == 2 &&
-    dest[ 0 ] == L'm' && iswdigit( dest[ 1 ] ) );
+  return tagName.size() == 2 && tagName[ 0 ] == L'm' && iswdigit( tagName[ 1 ] );
 }
 
-ArticleDom::ArticleDom( wstring const & str ):
+bool isAnyM( wstring const & tagName )
+{
+  return tagName == GD_NATIVE_TO_WS( L"m" ) || is_mN( tagName );
+}
+
+bool checkM( wstring const & dest, wstring const & src )
+{
+  return src == GD_NATIVE_TO_WS( L"m" ) && is_mN( dest );
+}
+
+/// Closing the [mN] tags is optional. Quote from https://documentation.help/ABBYY-Lingvo8/paragraph_form.htm:
+/// Any paragraph from this tag until the end of card or until system meets an «[/m]» (margin shift toggle off) tag
+struct MustTagBeClosed
+{
+  bool operator()( ArticleDom::Node const * tag ) const
+  {
+    Q_ASSERT( tag->isTag );
+    return !isAnyM( tag->tagName );
+  }
+};
+
+} // unnamed namespace
+
+ArticleDom::ArticleDom( wstring const & str, string const & dictName,
+                        wstring const & headword_):
   root( Node::Tag(), wstring(), wstring() ), stringPos( str.c_str() ),
-  transcriptionCount( 0 )
+  lineStartPos( str.c_str() ),
+  transcriptionCount( 0 ),
+  mediaCount( 0 ),
+  dictionaryName( dictName ),
+  headword( headword_ )
 {
   list< Node * > stack; // Currently opened tags
 
@@ -70,45 +214,155 @@ ArticleDom::ArticleDom( wstring const & str ):
     {
       nextChar();
 
+      if ( ch == L'@' && !escaped )
+      {
+        if( !atSignFirstInLine() )
+        {
+          // Not insided card
+          if( dictName.empty() )
+            gdWarning( "Unescaped '@' symbol found" );
+          else
+            gdWarning( "Unescaped '@' symbol found in \"%s\"", dictName.c_str() );
+        }
+        else
+        {
+          // Insided card
+          wstring linkTo;
+          nextChar();
+          for( ; ; nextChar() )
+          {
+            if( ch == L'\n' )
+              break;
+            if( ch != L'\r' )
+            {
+              if( escaped && ( ch == L'(' || ch == ')' ) )
+                linkTo.push_back( L'\\' );
+              linkTo.push_back( ch );
+            }
+          }
+          linkTo = Folding::trimWhitespace( linkTo );
+
+          if( !linkTo.empty() )
+          {
+            list< wstring > allLinkEntries;
+            processUnsortedParts( linkTo, true );
+            expandOptionalParts( linkTo, &allLinkEntries );
+
+            for( list< wstring >::iterator entry = allLinkEntries.begin();
+                 entry != allLinkEntries.end(); )
+            {
+              if ( !textNode )
+              {
+                Node text = Node( Node::Text(), wstring() );
+
+                if ( stack.empty() )
+                {
+                  root.push_back( text );
+                  stack.push_back( &root.back() );
+                }
+                else
+                {
+                  stack.back()->push_back( text );
+                  stack.push_back( &stack.back()->back() );
+                }
+
+                textNode = stack.back();
+              }
+              textNode->text.push_back( L'-' );
+              textNode->text.push_back( L' ' );
+
+              // Close the currently opened text node
+              stack.pop_back();
+              textNode = 0;
+
+              wstring linkText = Folding::trimWhitespace( *entry );
+              ArticleDom nodeDom( linkText, dictName, headword_ );
+
+              Node link( Node::Tag(), GD_NATIVE_TO_WS( L"@" ), wstring() );
+              for( Node::iterator n = nodeDom.root.begin(); n != nodeDom.root.end(); ++n )
+                link.push_back( *n );
+
+              ++entry;
+
+              if ( stack.empty() )
+              {
+                root.push_back( link );
+                if( entry != allLinkEntries.end() ) // Add line break before next entry
+                  root.push_back( Node( Node::Tag(), GD_NATIVE_TO_WS( L"br" ), wstring() ) );
+              }
+              else
+              {
+                stack.back()->push_back( link );
+                if( entry != allLinkEntries.end() )
+                  stack.back()->push_back( Node( Node::Tag(), GD_NATIVE_TO_WS( L"br" ), wstring() ) );
+              }
+            }
+
+            // Skip to next '@'
+
+            while( !( ch == L'@' && !escaped && atSignFirstInLine() ) )
+              nextChar();
+
+            stringPos--;
+            ch = L'\n';
+            escaped = false;
+          }
+        }
+      } // if ( ch == L'@' )
+
       if ( ch == L'[' && !escaped )
       {
         // Beginning of a tag.
-        do
-        {
-          nextChar();
-        } while( Folding::isWhitespace( ch ) );
-
         bool isClosing;
-
-        if ( ch == L'/' && !escaped )
-        {
-          // A closing tag.
-          isClosing = true;
-          nextChar();
-        }
-        else
-          isClosing = false;
-
-        // Read tag's name
         wstring name;
-
-        while( ( ch != L']' || escaped ) && !Folding::isWhitespace( ch ) )
-        {
-          name.push_back( ch );
-          nextChar();
-        }
-
-        while( Folding::isWhitespace( ch ) )
-          nextChar();
-
-        // Read attrs
-
         wstring attrs;
 
-        while( ch != L']' || escaped )
+        try
         {
-          attrs.push_back( ch );
-          nextChar();
+          do
+          {
+            nextChar();
+          } while( Folding::isWhitespace( ch ) );
+
+          if ( ch == L'/' && !escaped )
+          {
+            // A closing tag.
+            isClosing = true;
+            nextChar();
+          }
+          else
+            isClosing = false;
+
+          // Read tag's name
+
+          while( ( ch != L']' || escaped ) && !Folding::isWhitespace( ch ) )
+          {
+            name.push_back( ch );
+            nextChar();
+          }
+
+          while( Folding::isWhitespace( ch ) )
+            nextChar();
+
+          // Read attrs
+
+          while( ch != L']' || escaped )
+          {
+            attrs.push_back( ch );
+            nextChar();
+          }
+        }
+        catch( eot )
+        {
+          if( !dictionaryName.empty() )
+            gdWarning( "DSL: Unfinished tag \"%s\" with attributes \"%s\" found in \"%s\", article \"%s\".",
+                       gd::toQString( name ).toUtf8().data(), gd::toQString( attrs ).toUtf8().data(),
+                       dictionaryName.c_str(), gd::toQString( headword ).toUtf8().data() );
+          else
+            gdWarning( "DSL: Unfinished tag \"%s\" with attributes \"%s\" found",
+                       gd::toQString( name ).toUtf8().data(), gd::toQString( attrs ).toUtf8().data() );
+
+          throw eot();
         }
 
         // Add the tag, or close it
@@ -132,25 +386,30 @@ ArticleDom::ArticleDom( wstring const & str ):
             ++transcriptionCount;
         }
         
-        if ( !isClosing )
+        // If the tag is [s], we update the mediaCount
+        if ( name == GD_NATIVE_TO_WS( L"s" ) )
         {
-          if ( name.size() == 2 && name[ 0 ] == L'm' && iswdigit( name[ 1 ] ) )
+          if ( isClosing )
           {
-            // Opening an 'mX' tag closes any previous 'm' tag
-            closeTag( GD_NATIVE_TO_WS( L"m" ), stack, false );
-          }
-          
-          Node node( Node::Tag(), name, attrs );
-
-          if ( stack.empty() )
-          {
-            root.push_back( node );
-            stack.push_back( &root.back() );
+            if ( mediaCount )
+              --mediaCount;
           }
           else
+            ++mediaCount;
+        }
+
+        if ( !isClosing )
+        {
+          if( isAnyM( name ) )
           {
-            stack.back()->push_back( node );
-            stack.push_back( &stack.back()->back() );
+            // Opening an 'mX' or 'm' tag closes any previous 'm' tag
+            closeTag( GD_NATIVE_TO_WS( L"m" ), stack, false );
+          }
+          openTag( name, attrs, stack );
+          if ( name == GD_NATIVE_TO_WS( L"br" ) )
+          {
+            // [br] tag don't have closing tag
+            closeTag( name, stack );
           }
         }
         else
@@ -186,7 +445,7 @@ ArticleDom::ArticleDom( wstring const & str ):
             nextChar();
           } while( Folding::isWhitespace( ch ) );
 
-          wstring linkTo;
+          wstring linkTo, linkText;
 
           for( ; ; nextChar() )
           {
@@ -201,10 +460,21 @@ ArticleDom::ArticleDom( wstring const & str ):
               {
                 linkTo.push_back( L'>' );
                 linkTo.push_back( ch );
+
+                linkText.push_back( L'>' );
+                if( escaped )
+                  linkText.push_back( L'\\' );
+                linkText.push_back( ch );
               }
             }
             else
+            {
               linkTo.push_back( ch );
+
+              if( escaped )
+                linkText.push_back( L'\\' );
+              linkText.push_back( ch );
+            }
           }
 
           // Add the corresponding node
@@ -216,8 +486,13 @@ ArticleDom::ArticleDom( wstring const & str ):
             textNode = 0;
           }
 
+          linkText = Folding::trimWhitespace( linkText );
+          processUnsortedParts( linkText, true );
+          ArticleDom nodeDom( linkText, dictName, headword_ );
+
           Node link( Node::Tag(), GD_NATIVE_TO_WS( L"ref" ), wstring() );
-          link.push_back( Node( Node::Text(), linkTo ) );
+          for( Node::iterator n = nodeDom.root.begin(); n != nodeDom.root.end(); ++n )
+            link.push_back( *n );
 
           if ( stack.empty() )
             root.push_back( link );
@@ -365,7 +640,7 @@ ArticleDom::ArticleDom( wstring const & str ):
         }
       }
 
-      if ( escaped && ch == L' ' )
+      if ( escaped && ch == L' ' && mediaCount == 0 )
         ch = 0xA0; // Escaped spaces turn into non-breakable ones in Lingvo
             
       textNode->text.push_back( ch );
@@ -379,7 +654,92 @@ ArticleDom::ArticleDom( wstring const & str ):
     stack.pop_back();
 
   if ( stack.size() )
-    fprintf( stderr, "Warning: %u tags were unclosed.\n", stack.size() );
+  {
+    list< Node * >::iterator it = std::find_if( stack.begin(), stack.end(), MustTagBeClosed() );
+    if( it == stack.end() )
+      return; // no unclosed tags that must be closed => nothing to warn about
+    QByteArray const firstTagName = gd::toQString( ( *it )->tagName ).toUtf8();
+    ++it;
+    unsigned const unclosedTagCount = 1 + std::count_if( it, stack.end(), MustTagBeClosed() );
+
+    if( dictName.empty() )
+    {
+      gdWarning( "Warning: %u tag(s) were unclosed, first tag name \"%s\".\n",
+                 unclosedTagCount, firstTagName.constData() );
+    }
+    else
+    {
+      gdWarning( "Warning: %u tag(s) were unclosed in \"%s\", article \"%s\", first tag name \"%s\".\n",
+                 unclosedTagCount, dictName.c_str(), gd::toQString( headword ).toUtf8().constData(),
+                 firstTagName.constData() );
+    }
+  }
+}
+
+void ArticleDom::openTag( wstring const & name,
+                          wstring const & attrs,
+                          list<Node *> &stack )
+{
+  list< Node > nodesToReopen;
+
+  if( isAnyM( name ) )
+  {
+    // All tags above [m] tag will be closed and reopened after
+    // to avoid break this tag by closing some other tag.
+
+    while( stack.size() )
+    {
+      nodesToReopen.push_back( Node( Node::Tag(), stack.back()->tagName,
+                                     stack.back()->tagAttrs ) );
+
+      if ( stack.back()->empty() )
+      {
+        // Empty nodes are deleted since they're no use
+
+        stack.pop_back();
+
+        Node * parent = stack.size() ? stack.back() : &root;
+
+        parent->pop_back();
+      }
+      else
+        stack.pop_back();
+    }
+  }
+
+  // Add tag
+
+  Node node( Node::Tag(), name, attrs );
+
+  if ( stack.empty() )
+  {
+    root.push_back( node );
+    stack.push_back( &root.back() );
+  }
+  else
+  {
+    stack.back()->push_back( node );
+    stack.push_back( &stack.back()->back() );
+  }
+
+  // Reopen tags if needed
+
+  while( nodesToReopen.size() )
+  {
+    if ( stack.empty() )
+    {
+      root.push_back( nodesToReopen.back() );
+      stack.push_back( &root.back() );
+    }
+    else
+    {
+      stack.back()->push_back( nodesToReopen.back() );
+      stack.push_back( &stack.back()->back() );
+    }
+
+    nodesToReopen.pop_back();
+  }
+
 }
 
 void ArticleDom::closeTag( wstring const & name,
@@ -416,9 +776,9 @@ void ArticleDom::closeTag( wstring const & name,
         nodesToReopen.push_back( Node( Node::Tag(), stack.back()->tagName,
                                        stack.back()->tagAttrs ) );
 
-      if ( stack.back()->empty() )
+      if ( stack.back()->empty() && stack.back()->tagName != GD_NATIVE_TO_WS( L"br" ) )
       {
-        // Empty nodes are deleted since they're no use
+        // Empty nodes except [br] tag are deleted since they're no use
 
         stack.pop_back();
 
@@ -452,12 +812,17 @@ void ArticleDom::closeTag( wstring const & name,
   else
   if ( warn )
   {
-    fprintf( stderr, "Warning: no corresponding opening tag for closing tag \"/%ls\" found.\n",
-             name.c_str() );
+    if( !dictionaryName.empty() )
+      gdWarning( "No corresponding opening tag for closing tag \"%s\" found in \"%s\", article \"%s\".",
+                 gd::toQString( name ).toUtf8().data(), dictionaryName.c_str(),
+                 gd::toQString( headword ).toUtf8().data() );
+    else
+      gdWarning( "No corresponding opening tag for closing tag \"%s\" found.",
+                 gd::toQString( name ).toUtf8().data() );
   }
 }
 
-void ArticleDom::nextChar() throw( eot )
+void ArticleDom::nextChar() THROW_SPEC( eot )
 {
   if ( !*stringPos )
     throw eot();
@@ -487,19 +852,30 @@ void ArticleDom::nextChar() throw( eot )
   }
   else
     escaped = false;
+
+  if( ch == '\n' || ch == '\r' )
+    lineStartPos = stringPos;
 }
 
+bool ArticleDom::atSignFirstInLine()
+{
+  // Check if '@' sign is first after '\n', leading spaces and dsl tags
+  if( stringPos <= lineStartPos )
+    return true;
+
+  return isAtSignFirst( wstring( lineStartPos ) );
+}
 
 /////////////// DslScanner
 
-DslScanner::DslScanner( string const & fileName ) throw( Ex, Iconv::Ex ):
+DslScanner::DslScanner( string const & fileName ) THROW_SPEC( Ex, Iconv::Ex ):
   encoding( Windows1252 ), iconv( encoding ), readBufferPtr( readBuffer ),
-  readBufferLeft( 0 ), linesRead( 0 )
+  readBufferLeft( 0 ), wcharBuffer( 64 ), linesRead( 0 )
 {
   // Since .dz is backwards-compatible with .gz, we use gz- functions to
   // read it -- they are much nicer than the dict_data- ones.
-  f = gzopen( fileName.c_str(), "rb");
 
+  f = gd_gzopen( fileName.c_str() );
   if ( !f )
     throw exCantOpen( fileName );
 
@@ -581,6 +957,7 @@ DslScanner::DslScanner( string const & fileName ) throw( Ex, Iconv::Ex ):
     bool isName = false;
     bool isLangFrom = false;
     bool isLangTo = false;
+    bool isSoundDict = false;
 
     if ( !str.compare( 0, 5, GD_NATIVE_TO_WS( L"#NAME" ), 5 ) )
       isName = true;
@@ -591,6 +968,9 @@ DslScanner::DslScanner( string const & fileName ) throw( Ex, Iconv::Ex ):
     if ( !str.compare( 0, 18, GD_NATIVE_TO_WS( L"#CONTENTS_LANGUAGE" ), 18 ) )
       isLangTo = true;
     else
+    if ( !str.compare( 0, 17, GD_NATIVE_TO_WS( L"#SOUND_DICTIONARY" ), 17 ) )
+      isSoundDict = true;
+    else
     if ( str.compare( 0, 17, GD_NATIVE_TO_WS( L"#SOURCE_CODE_PAGE" ), 17 ) )
       continue;
 
@@ -614,13 +994,15 @@ DslScanner::DslScanner( string const & fileName ) throw( Ex, Iconv::Ex ):
       langFrom = arg;
     else if ( isLangTo )
       langTo = arg;
+    else if ( isSoundDict )
+      soundDictionary = arg;
     else
     {
       // The encoding
       if ( !needExactEncoding )
       {
         // We don't need that!
-        fprintf( stderr, "Warning: encoding was specified in a Unicode file, ignoring.\n" );
+        GD_FDPRINTF( stderr, "Warning: encoding was specified in a Unicode file, ignoring.\n" );
       }
       else
       if ( !wcscasecmp( arg.c_str(), GD_NATIVE_TO_WS( L"Latin" ) ) )
@@ -643,6 +1025,8 @@ DslScanner::DslScanner( string const & fileName ) throw( Ex, Iconv::Ex ):
   // We need to rewind to that line so readNextLine() would return it again
   // next time it's called. To do that, we just use the slow gzseek() and
   // empty the read buffer.
+  if( gzdirect( f ) )                    // Without this ZLib 1.2.7 gzread() return 0
+    gzrewind( f );                       // after gzseek() call on uncompressed files
   gzseek( f, offset, SEEK_SET );
   readBufferPtr = readBuffer;
   readBufferLeft = 0;
@@ -656,7 +1040,7 @@ DslScanner::~DslScanner() throw()
   gzclose( f );
 }
 
-bool DslScanner::readNextLine( wstring & out, size_t & offset ) throw( Ex,
+bool DslScanner::readNextLine( wstring & out, size_t & offset ) THROW_SPEC( Ex,
                                                                        Iconv::Ex )
 {
   offset = (size_t)( gztell( f ) - readBufferLeft );
@@ -738,7 +1122,7 @@ bool DslScanner::readNextLine( wstring & out, size_t & offset ) throw( Ex,
       --leftInOut; // Complements the next decremention
     }
 
-    if ( ( r != Iconv::Success && r != Iconv::NeedMoreOut ) || outBytesLeft )
+    if ( outBytesLeft )
       throw exEncodingError();
 
     --leftInOut;
@@ -761,15 +1145,43 @@ bool DslScanner::readNextLine( wstring & out, size_t & offset ) throw( Ex,
   }
 }
 
+bool DslScanner::readNextLineWithoutComments( wstring & out, size_t & offset )
+                 THROW_SPEC( Ex, Iconv::Ex )
+{
+  wstring str;
+  bool commentToNextLine = false;
+  size_t currentOffset;
+
+  out.erase();
+  offset = 0;
+
+  do
+  {
+    bool b = readNextLine( str, currentOffset );
+
+    if( offset == 0 )
+      offset = currentOffset;
+
+    if( !b )
+      return false;
+
+    stripComments( str, commentToNextLine);
+
+    out += str;
+  }
+  while( commentToNextLine );
+
+  return true;
+}
 
 /////////////// DslScanner
 
-DslIconv::DslIconv( DslEncoding e ) throw( Iconv::Ex ):
+DslIconv::DslIconv( DslEncoding e ) THROW_SPEC( Iconv::Ex ):
   Iconv( Iconv::GdWchar, getEncodingNameFor( e ) )
 {
 }
 
-void DslIconv::reinit( DslEncoding e ) throw( Iconv::Ex )
+void DslIconv::reinit( DslEncoding e ) THROW_SPEC( Iconv::Ex )
 {
   Iconv::reinit( Iconv::GdWchar, getEncodingNameFor( e ) );
 }
@@ -837,7 +1249,7 @@ void processUnsortedParts( wstring & str, bool strip )
 
       if ( refCount < 0 )
       {
-        fprintf( stderr, "Warning: an unmatched closing brace was encountered.\n" );
+        GD_FDPRINTF( stderr, "Warning: an unmatched closing brace was encountered.\n" );
         refCount = 0;
         // But we remove that thing either way
         str.erase( x, 1 );
@@ -866,14 +1278,25 @@ void processUnsortedParts( wstring & str, bool strip )
 
   if ( strip && refCount )
   {
-    fprintf( stderr, "Warning: unclosed brace(s) encountered.\n" );
+    GD_FDPRINTF( stderr, "Warning: unclosed brace(s) encountered.\n" );
     str.erase( startPos );
   }
 }
 
-void expandOptionalParts( wstring & str, list< wstring > & result,
-                          size_t x )
+void expandOptionalParts( wstring & str, list< wstring > * result,
+                          size_t x, bool inside_recurse )
 {
+  if( str.size() > 500 )
+  {
+    // Don't expand too long headwords - it is highly likely incorrect dictionary
+    result->push_back( str );
+    return;
+  }
+
+  list< wstring > expanded;
+  list< wstring > * headwords;
+  headwords = inside_recurse ? result : &expanded;
+
   for( ; x < str.size(); )
   {
     wchar ch = str[ x ];
@@ -916,7 +1339,7 @@ void expandOptionalParts( wstring & str, list< wstring > & result,
                 wstring removed( str, 0, x );
                 removed.append( str, y + 1, str.size() - y - 1 );
 
-                expandOptionalParts( removed, result, x );
+                expandOptionalParts( removed, headwords, x, true );
               }
 
               break;
@@ -931,10 +1354,14 @@ void expandOptionalParts( wstring & str, list< wstring > & result,
           wstring removed( str, 0, x );
 
           // Limit the amount of results to avoid excessive resource consumption
-          if ( result.size() < 32 )
-            result.push_back( removed );
+          if ( headwords->size() < 32 )
+            headwords->push_back( removed );
           else
+          {
+            if( !inside_recurse )
+              result->merge( expanded );
             return;
+          }
         }
       }
 
@@ -954,20 +1381,62 @@ void expandOptionalParts( wstring & str, list< wstring > & result,
   }
 
   // Limit the amount of results to avoid excessive resource consumption
-  if ( result.size() < 32 )
-    result.push_back( str );
+  if ( headwords->size() < 32 )
+    headwords->push_back( str );
+  if( !inside_recurse )
+    result->merge( expanded );
+}
+
+static const wstring openBraces( GD_NATIVE_TO_WS( L"{{" ) );
+static const wstring closeBraces( GD_NATIVE_TO_WS( L"}}" ) );
+
+void stripComments( wstring & str, bool & nextLine )
+{
+  string::size_type n = 0, n2 = 0;
+
+  for( ; ; )
+  {
+    if( nextLine )
+    {
+      n = str.find( closeBraces, n2 );
+      if( n == string::npos )
+      {
+        str.erase( n2, n );
+        return;
+      }
+      str.erase( n2, n - n2 + 2 );
+      nextLine = false;
+    }
+
+    n = str.find( openBraces, n2 );
+    if( n == string::npos )
+      return;
+    nextLine = true;
+    n2 = n;
+  }
 }
 
 void expandTildes( wstring & str, wstring const & tildeReplacement )
 {
+  wstring tildeValue = Folding::trimWhitespace( tildeReplacement );
   for( size_t x = 0; x < str.size(); )
     if ( str[ x ] == L'\\' )
       x+=2;
     else
     if ( str[ x ] == L'~' )
     {
-      str.replace( x, 1, tildeReplacement );
-      x += tildeReplacement.size();
+      if( x > 0 && str[ x - 1 ] == '^' && ( x < 2 || str[ x - 2 ] != '\\' ) )
+      {
+        str.replace( x - 1, 2, tildeValue );
+        str[ x - 1 ] = QChar( str[ x - 1 ] ).isUpper() ? QChar::toLower( (uint)str[ x - 1 ] )
+                                                       : QChar::toUpper( (uint)str[ x - 1 ] );
+        x = x - 1 + tildeValue.size();
+      }
+      else
+      {
+        str.replace( x, 1, tildeValue );
+        x += tildeValue.size();
+      }
     }
     else
       ++x;
@@ -999,6 +1468,10 @@ void normalizeHeadword( wstring & str )
       }
     }
   }
+  if( !str.empty() && str[ str.size() - 1 ] == L' ' )
+    str.erase( str.size() - 1, 1 );
+  if( !str.empty() && str[ 0 ] == L' ' )
+    str.erase( 0, 1 );
 }
 
 namespace
