@@ -1,31 +1,33 @@
-/* This file is (c) 2008-2011 Konstantin Isakov <ikm@users.berlios.de>
+/* This file is (c) 2008-2012 Konstantin Isakov <ikm@goldendict.org>
  * Part of GoldenDict. Licensed under GPLv3 or later, see the LICENSE file */
 
 #include "folding.hh"
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+#include <QRegularExpression>
+#else
+#include <QRegExp>
+#endif
 
 namespace Folding {
 
-namespace
-{
-  #include "inc_case_folding.hh"
-  #include "inc_diacritic_folding.hh"
+#include "inc_case_folding.hh"
+#include "inc_diacritic_folding.hh"
 
-  /// Tests if the given char is one of the Unicode combining marks. Some are
-  /// caught by the diacritics folding table, but they are only handled there
-  /// when they come with their main characters, not by themselves. The rest
-  /// are caught here.
-  bool isCombiningMark( wchar ch )
-  {
-    return (
-             ( ch >= 0x300 && ch <= 0x36F ) ||
-             ( ch >= 0x1DC0 && ch <= 0x1DFF ) ||
-             ( ch >= 0x20D0 && ch <= 0x20FF ) ||
-             ( ch >= 0xFE20 && ch <= 0xFE2F )
-           );
-  }
+/// Tests if the given char is one of the Unicode combining marks. Some are
+/// caught by the diacritics folding table, but they are only handled there
+/// when they come with their main characters, not by themselves. The rest
+/// are caught here.
+bool isCombiningMark( wchar ch )
+{
+  return (
+           ( ch >= 0x300 && ch <= 0x36F ) ||
+           ( ch >= 0x1DC0 && ch <= 0x1DFF ) ||
+           ( ch >= 0x20D0 && ch <= 0x20FF ) ||
+           ( ch >= 0xFE20 && ch <= 0xFE2F )
+         );
 }
 
-wstring apply( wstring const & in )
+wstring apply( wstring const & in, bool preserveWildcards )
 {
   // First, strip diacritics and apply ws/punctuation removal
 
@@ -41,7 +43,12 @@ wstring apply( wstring const & in )
   {
     wchar ch = foldDiacritic( nextChar, left, consumed );
 
-    if ( !isCombiningMark( ch ) && !isWhitespace( ch ) && !isPunct( ch ) )
+    if ( !isCombiningMark( ch ) && !isWhitespace( ch )
+         && ( !isPunct( ch )
+              || ( preserveWildcards &&
+                   ( ch == '\\' || ch == '?' || ch == '*' || ch == '[' || ch == ']' ) )
+            )
+    )
       withoutDiacritics.push_back( ch );
 
     nextChar += consumed;
@@ -659,4 +666,32 @@ void normalizeWhitespace( wstring & str )
   }
 }
 
+QString escapeWildcardSymbols( const QString & str )
+{
+  QString escaped( str );
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+  escaped.replace( QRegularExpression( "([\\[\\]\\?\\*])" ), "\\\\1" );
+#else
+  escaped.replace( QRegExp( "([\\[\\]\\?\\*])", Qt::CaseInsensitive ), "\\\\1" );
+#endif
+  return escaped;
+}
+
+QString unescapeWildcardSymbols( const QString & str )
+{
+  QString unescaped( str );
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 0, 0 )
+  unescaped.replace( QRegularExpression( "\\\\([\\[\\]\\?\\*])" ), "\\1" );
+#else
+  unescaped.replace( QRegExp( "\\\\([\\[\\]\\?\\*])", Qt::CaseInsensitive ), "\\1" );
+#endif
+  return unescaped;
+}
+
+wchar foldedDiacritic( wchar const * in, size_t size, size_t & consumed )
+{
+  return foldDiacritic( in, size, consumed );
+}
+
+
 }
