@@ -1,39 +1,89 @@
-/* This file is (c) 2008-2011 Konstantin Isakov <ikm@users.berlios.de>
+/* This file is (c) 2008-2012 Konstantin Isakov <ikm@goldendict.org>
  * Part of GoldenDict. Licensed under GPLv3 or later, see the LICENSE file */
 
 #include "sources.hh"
 #include <QFileDialog>
 #include <QMessageBox>
-#include <QCryptographicHash>
-#include <QDateTime>
 #include <QStandardItemModel>
+#include "gddebug.hh"
 
-Sources::Sources( QWidget * parent, Config::Paths const & paths,
-                  Config::SoundDirs const & soundDirs,
-                  Config::Hunspell const & hunspell,
-                  Config::Transliteration const & trs,
-                  Config::Forvo const & forvo,
-                  Config::MediaWikis const & mediawikis,
-                  Config::WebSites const & webSites ): QWidget( parent ),
-  mediawikisModel( this, mediawikis ),
-  webSitesModel( this, webSites ),
-  pathsModel( this, paths ),
-  soundDirsModel( this, soundDirs ),
-  hunspellDictsModel( this, hunspell )
+#ifdef MAKE_CHINESE_CONVERSION_SUPPORT
+#include "chineseconversion.hh"
+#endif
+
+
+Sources::Sources( QWidget * parent, Config::Class const & cfg):
+  QWidget( parent ),
+#ifdef MAKE_CHINESE_CONVERSION_SUPPORT
+  chineseConversion( new ChineseConversion( this, cfg.transliteration.chinese ) ),
+#endif
+#if defined( Q_OS_WIN32 ) || defined( Q_OS_MAC )
+  textToSpeechSource( NULL ),
+#endif
+  itemDelegate( new QItemDelegate( this ) ),
+  itemEditorFactory( new QItemEditorFactory() ),
+  mediawikisModel( this, cfg.mediawikis ),
+  webSitesModel( this, cfg.webSites ),
+  dictServersModel( this, cfg.dictServers ),
+  programsModel( this, cfg.programs ),
+  pathsModel( this, cfg.paths ),
+  soundDirsModel( this, cfg.soundDirs ),
+  hunspellDictsModel( this, cfg.hunspell )
 {
   ui.setupUi( this );
 
+  Config::Hunspell const & hunspell = cfg.hunspell;
+  Config::Transliteration const & trs = cfg.transliteration;
+  Config::Forvo const & forvo = cfg.forvo;
+
+  // TODO: will programTypeEditorCreator and itemEditorFactory be destroyed by
+  // anyone?
+  QItemEditorCreatorBase * programTypeEditorCreator =
+         new QStandardItemEditorCreator< ProgramTypeEditor >();
+
+  itemEditorFactory->registerEditor( QVariant::Int, programTypeEditorCreator );
+
+  itemDelegate->setItemEditorFactory( itemEditorFactory );
+
   ui.mediaWikis->setTabKeyNavigation( true );
   ui.mediaWikis->setModel( &mediawikisModel );
   ui.mediaWikis->resizeColumnToContents( 0 );
   ui.mediaWikis->resizeColumnToContents( 1 );
   ui.mediaWikis->resizeColumnToContents( 2 );
+  ui.mediaWikis->resizeColumnToContents( 3 );
 
   ui.webSites->setTabKeyNavigation( true );
   ui.webSites->setModel( &webSitesModel );
   ui.webSites->resizeColumnToContents( 0 );
   ui.webSites->resizeColumnToContents( 1 );
   ui.webSites->resizeColumnToContents( 2 );
+  ui.webSites->resizeColumnToContents( 3 );
+  ui.webSites->resizeColumnToContents( 4 );
+
+  ui.dictServers->setTabKeyNavigation( true );
+  ui.dictServers->setModel( &dictServersModel );
+  ui.dictServers->resizeColumnToContents( 0 );
+  ui.dictServers->resizeColumnToContents( 1 );
+  ui.dictServers->resizeColumnToContents( 2 );
+  ui.dictServers->resizeColumnToContents( 3 );
+  ui.dictServers->resizeColumnToContents( 4 );
+  ui.dictServers->resizeColumnToContents( 5 );
+
+  ui.programs->setTabKeyNavigation( true );
+  ui.programs->setModel( &programsModel );
+  ui.programs->resizeColumnToContents( 0 );
+  // Make sure this thing will be large enough
+  ui.programs->setColumnWidth( 1,
+#if QT_VERSION >= QT_VERSION_CHECK( 5, 11, 0 )
+    QFontMetrics( QFont() ).horizontalAdvance(
+#else
+    QFontMetrics( QFont() ).width(
+#endif
+      ProgramTypeEditor::getNameForType( Config::Program::PrefixMatch ) ) + 16 );
+  ui.programs->resizeColumnToContents( 2 );
+  ui.programs->resizeColumnToContents( 3 );
+  ui.programs->resizeColumnToContents( 4 );
+  ui.programs->setItemDelegate( itemDelegate );
 
   ui.paths->setTabKeyNavigation( true );
   ui.paths->setModel( &pathsModel );
@@ -54,6 +104,13 @@ Sources::Sources( QWidget * parent, Config::Paths const & paths,
   ui.enableRussianTransliteration->setChecked( trs.enableRussianTransliteration );
   ui.enableGermanTransliteration->setChecked( trs.enableGermanTransliteration );
   ui.enableGreekTransliteration->setChecked( trs.enableGreekTransliteration );
+  ui.enableBelarusianTransliteration->setChecked( trs.enableBelarusianTransliteration );
+
+#ifdef MAKE_CHINESE_CONVERSION_SUPPORT
+  ui.transliterationLayout->addWidget(chineseConversion);
+  ui.transliterationLayout->addItem(new QSpacerItem(20, 40, QSizePolicy::Minimum, QSizePolicy::Expanding));
+#endif
+
   ui.enableRomaji->setChecked( trs.romaji.enable );
   ui.enableHepburn->setChecked( trs.romaji.enableHepburn );
   ui.enableNihonShiki->setChecked( trs.romaji.enableNihonShiki );
@@ -65,6 +122,12 @@ Sources::Sources( QWidget * parent, Config::Paths const & paths,
   ui.forvoApiKey->setText( forvo.apiKey );
   ui.forvoLanguageCodes->setText( forvo.languageCodes );
 
+  // Text to speech
+#if defined( Q_OS_WIN32 ) || defined( Q_OS_MAC )
+  textToSpeechSource = new TextToSpeechSource( this, cfg.voiceEngines );
+  ui.tabWidget->addTab( textToSpeechSource, QIcon(":/icons/playsound_color.png"), tr( "Text to Speech" ) );
+#endif
+
   if ( Config::isPortableVersion() )
   {
     // Paths
@@ -103,6 +166,7 @@ void Sources::fitSoundDirsColumns()
 {
   ui.soundDirs->resizeColumnToContents( 0 );
   ui.soundDirs->resizeColumnToContents( 1 );
+  ui.soundDirs->resizeColumnToContents( 2 );
 }
 
 void Sources::fitHunspellDictsColumns()
@@ -226,6 +290,65 @@ void Sources::on_removeWebSite_clicked()
     webSitesModel.removeSite( current.row() );
 }
 
+void Sources::on_addDictServer_clicked()
+{
+  dictServersModel.addNewServer();
+
+  QModelIndex result =
+    dictServersModel.index( dictServersModel.rowCount( QModelIndex() ) - 1,
+                            1, QModelIndex() );
+
+  ui.dictServers->scrollTo( result );
+  ui.dictServers->edit( result );
+}
+
+void Sources::on_removeDictServer_clicked()
+{
+  QModelIndex current = ui.dictServers->currentIndex();
+
+  if ( current.isValid() &&
+       QMessageBox::question( this, tr( "Confirm removal" ),
+                              tr( "Remove site <b>%1</b> from the list?" ).arg( dictServersModel.getCurrentDictServers()[ current.row() ].name ),
+                              QMessageBox::Ok,
+                              QMessageBox::Cancel ) == QMessageBox::Ok )
+    dictServersModel.removeServer( current.row() );
+}
+
+void Sources::on_addProgram_clicked()
+{
+  programsModel.addNewProgram();
+
+  QModelIndex result =
+    programsModel.index( programsModel.rowCount( QModelIndex() ) - 1,
+                         1, QModelIndex() );
+
+  ui.programs->scrollTo( result );
+  ui.programs->edit( result );
+}
+
+void Sources::on_removeProgram_clicked()
+{
+  QModelIndex current = ui.programs->currentIndex();
+
+  if ( current.isValid() &&
+       QMessageBox::question( this, tr( "Confirm removal" ),
+                              tr( "Remove program <b>%1</b> from the list?" ).arg( programsModel.getCurrentPrograms()[ current.row() ].name ),
+                              QMessageBox::Ok,
+                              QMessageBox::Cancel ) == QMessageBox::Ok )
+    programsModel.removeProgram( current.row() );
+}
+
+Config::VoiceEngines Sources::getVoiceEngines() const
+{
+#if defined( Q_OS_WIN32 ) || defined( Q_OS_MAC )
+  if ( !textToSpeechSource )
+    return Config::VoiceEngines();
+  return textToSpeechSource->getVoiceEnginesModel().getCurrentVoiceEngines();
+#else
+  return Config::VoiceEngines();
+#endif
+}
+
 Config::Hunspell Sources::getHunspell() const
 {
   Config::Hunspell h;
@@ -243,6 +366,10 @@ Config::Transliteration Sources::getTransliteration() const
   tr.enableRussianTransliteration = ui.enableRussianTransliteration->isChecked();
   tr.enableGermanTransliteration = ui.enableGermanTransliteration->isChecked();
   tr.enableGreekTransliteration = ui.enableGreekTransliteration->isChecked();
+  tr.enableBelarusianTransliteration = ui.enableBelarusianTransliteration->isChecked();
+#ifdef MAKE_CHINESE_CONVERSION_SUPPORT
+  chineseConversion->getConfig( tr.chinese );
+#endif
   tr.romaji.enable = ui.enableRomaji->isChecked();
   tr.romaji.enableHepburn = ui.enableHepburn->isChecked();
   tr.romaji.enableNihonShiki = ui.enableNihonShiki->isChecked();
@@ -284,11 +411,7 @@ void MediaWikisModel::addNewWiki()
 
   w.enabled = false;
 
-  // That's quite some rng
-  w.id = QString(
-    QCryptographicHash::hash(
-      QDateTime::currentDateTime().toString( "\"MediaWiki\"dd.MM.yyyy hh:mm:ss.zzz" ).toUtf8(),
-      QCryptographicHash::Md5 ).toHex() );
+  w.id = Dictionary::generateRandomDictionaryId();
 
   w.url = "http://";
 
@@ -299,7 +422,7 @@ void MediaWikisModel::addNewWiki()
 
 QModelIndex MediaWikisModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
 {
-  return createIndex( row, column, 0 );
+  return createIndex( row, column );
 }
 
 QModelIndex MediaWikisModel::parent( QModelIndex const & /*parent*/ ) const
@@ -335,7 +458,7 @@ int MediaWikisModel::columnCount( QModelIndex const & parent ) const
   if ( parent.isValid() )
     return 0;
   else
-    return 3;
+    return 4;
 }
 
 QVariant MediaWikisModel::headerData( int section, Qt::Orientation /*orientation*/, int role ) const
@@ -349,6 +472,8 @@ QVariant MediaWikisModel::headerData( int section, Qt::Orientation /*orientation
         return tr( "Name" );
       case 2:
         return tr( "Address" );
+      case 3:
+        return tr( "Icon" );
       default:
         return QVariant();
     }
@@ -358,7 +483,7 @@ QVariant MediaWikisModel::headerData( int section, Qt::Orientation /*orientation
 
 QVariant MediaWikisModel::data( QModelIndex const & index, int role ) const
 {
-  if ( (unsigned)index.row() >= mediawikis.size() )
+  if ( index.row() >= mediawikis.size() )
     return QVariant();
 
   if ( role == Qt::DisplayRole || role == Qt::EditRole )
@@ -369,13 +494,15 @@ QVariant MediaWikisModel::data( QModelIndex const & index, int role ) const
         return mediawikis[ index.row() ].name;
       case 2:
         return mediawikis[ index.row() ].url;
+      case 3:
+        return mediawikis[ index.row() ].icon;
       default:
         return QVariant();
     }
   }
 
   if ( role == Qt::CheckStateRole && !index.column() )
-    return mediawikis[ index.row() ].enabled;
+    return mediawikis[ index.row() ].enabled ? Qt::Checked : Qt::Unchecked;
 
   return QVariant();
 }
@@ -383,13 +510,13 @@ QVariant MediaWikisModel::data( QModelIndex const & index, int role ) const
 bool MediaWikisModel::setData( QModelIndex const & index, const QVariant & value,
                                int role )
 {
-  if ( (unsigned)index.row() >= mediawikis.size() )
+  if ( index.row() >= mediawikis.size() )
     return false;
 
   if ( role == Qt::CheckStateRole && !index.column() )
   {
-    //printf( "type = %d\n", (int)value.type() );
-    //printf( "value = %d\n", (int)value.toInt() );
+    //DPRINTF( "type = %d\n", (int)value.type() );
+    //DPRINTF( "value = %d\n", (int)value.toInt() );
 
     // XXX it seems to be always passing Int( 2 ) as a value, so we just toggle
     mediawikis[ index.row() ].enabled = !mediawikis[ index.row() ].enabled;
@@ -409,6 +536,10 @@ bool MediaWikisModel::setData( QModelIndex const & index, const QVariant & value
         mediawikis[ index.row() ].url =  value.toString();
         dataChanged( index, index );
         return true;
+      case 3:
+        mediawikis[ index.row() ].icon =  value.toString();
+        dataChanged( index, index );
+        return true;
       default:
         return false;
     }
@@ -437,14 +568,12 @@ void WebSitesModel::addNewSite()
 
   w.enabled = false;
 
-  // That's quite some rng
-  w.id = QString(
-    QCryptographicHash::hash(
-      QDateTime::currentDateTime().toString( "\"WebSite\"dd.MM.yyyy hh:mm:ss.zzz" ).toUtf8(),
-      QCryptographicHash::Md5 ).toHex() );
+  w.id = Dictionary::generateRandomDictionaryId();
 
   w.url = "http://";
 
+  w.inside_iframe = true;
+
   beginInsertRows( QModelIndex(), webSites.size(), webSites.size() );
   webSites.push_back( w );
   endInsertRows();
@@ -452,7 +581,7 @@ void WebSitesModel::addNewSite()
 
 QModelIndex WebSitesModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
 {
-  return createIndex( row, column, 0 );
+  return createIndex( row, column );
 }
 
 QModelIndex WebSitesModel::parent( QModelIndex const & /*parent*/ ) const
@@ -466,7 +595,7 @@ Qt::ItemFlags WebSitesModel::flags( QModelIndex const & index ) const
 
   if ( index.isValid() )
   {
-    if ( !index.column() )
+    if ( index.column() <= 1 )
       result |= Qt::ItemIsUserCheckable;
     else
       result |= Qt::ItemIsEditable;
@@ -488,10 +617,195 @@ int WebSitesModel::columnCount( QModelIndex const & parent ) const
   if ( parent.isValid() )
     return 0;
   else
-    return 3;
+    return 5;
 }
 
 QVariant WebSitesModel::headerData( int section, Qt::Orientation /*orientation*/, int role ) const
+{
+  if( role == Qt::ToolTipRole )
+  {
+    if( section == 1 )
+      return tr( "Insert article as link inside <iframe> tag" );
+
+    return QVariant();
+  }
+
+  if ( role == Qt::DisplayRole )
+    switch( section )
+    {
+      case 0:
+        return tr( "Enabled" );
+      case 1:
+        return tr( "As link" );
+      case 2:
+        return tr( "Name" );
+      case 3:
+        return tr( "Address" );
+      case 4:
+        return tr( "Icon" );
+      default:
+        return QVariant();
+    }
+
+  return QVariant();
+}
+
+QVariant WebSitesModel::data( QModelIndex const & index, int role ) const
+{
+  if ( index.row() >= webSites.size() )
+    return QVariant();
+
+  if( role == Qt::ToolTipRole )
+  {
+    if( index.column() == 1 )
+      return tr( "Insert article as link inside <iframe> tag" );
+
+    return QVariant();
+  }
+
+  if ( role == Qt::DisplayRole || role == Qt::EditRole )
+  {
+    switch( index.column() )
+    {
+      case 2:
+        return webSites[ index.row() ].name;
+      case 3:
+        return webSites[ index.row() ].url;
+      case 4:
+        return webSites[ index.row() ].iconFilename;
+      default:
+        return QVariant();
+    }
+  }
+
+  if ( role == Qt::CheckStateRole && !index.column() )
+    return webSites[ index.row() ].enabled ? Qt::Checked : Qt::Unchecked;
+
+  if ( role == Qt::CheckStateRole && index.column() == 1 )
+    return webSites[ index.row() ].inside_iframe ? Qt::Checked : Qt::Unchecked;
+
+  return QVariant();
+}
+
+bool WebSitesModel::setData( QModelIndex const & index, const QVariant & value,
+                               int role )
+{
+  if ( index.row() >= webSites.size() )
+    return false;
+
+  if ( role == Qt::CheckStateRole && !index.column() )
+  {
+    //DPRINTF( "type = %d\n", (int)value.type() );
+    //DPRINTF( "value = %d\n", (int)value.toInt() );
+
+    // XXX it seems to be always passing Int( 2 ) as a value, so we just toggle
+    webSites[ index.row() ].enabled = !webSites[ index.row() ].enabled;
+
+    dataChanged( index, index );
+    return true;
+  }
+
+  if ( role == Qt::CheckStateRole && index.column() == 1 )
+  {
+    webSites[ index.row() ].inside_iframe = !webSites[ index.row() ].inside_iframe;
+
+    dataChanged( index, index );
+    return true;
+  }
+
+  if ( role == Qt::DisplayRole || role == Qt::EditRole )
+    switch( index.column() )
+    {
+      case 2:
+        webSites[ index.row() ].name =  value.toString();
+        dataChanged( index, index );
+        return true;
+      case 3:
+        webSites[ index.row() ].url =  value.toString();
+        dataChanged( index, index );
+        return true;
+      case 4:
+        webSites[ index.row() ].iconFilename =  value.toString();
+        dataChanged( index, index );
+        return true;
+      default:
+        return false;
+    }
+
+  return false;
+}
+
+////////// DictServersModel
+
+DictServersModel::DictServersModel( QWidget * parent,
+                              Config::DictServers const & dictServers_ ):
+  QAbstractItemModel( parent ), dictServers( dictServers_ )
+{
+}
+void DictServersModel::removeServer( int index )
+{
+  beginRemoveRows( QModelIndex(), index, index );
+  dictServers.erase( dictServers.begin() + index );
+  endRemoveRows();
+}
+
+void DictServersModel::addNewServer()
+{
+  Config::DictServer d;
+
+  d.enabled = false;
+
+  d.id = Dictionary::generateRandomDictionaryId();
+
+  d.url = "dict://";
+
+  beginInsertRows( QModelIndex(), dictServers.size(), dictServers.size() );
+  dictServers.push_back( d );
+  endInsertRows();
+}
+
+QModelIndex DictServersModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
+{
+  return createIndex( row, column );
+}
+
+QModelIndex DictServersModel::parent( QModelIndex const & /*parent*/ ) const
+{
+  return QModelIndex();
+}
+
+Qt::ItemFlags DictServersModel::flags( QModelIndex const & index ) const
+{
+  Qt::ItemFlags result = QAbstractItemModel::flags( index );
+
+  if ( index.isValid() )
+  {
+    if ( !index.column() )
+      result |= Qt::ItemIsUserCheckable;
+    else
+      result |= Qt::ItemIsEditable;
+  }
+
+  return result;
+}
+
+int DictServersModel::rowCount( QModelIndex const & parent ) const
+{
+  if ( parent.isValid() )
+    return 0;
+  else
+    return dictServers.size();
+}
+
+int DictServersModel::columnCount( QModelIndex const & parent ) const
+{
+  if ( parent.isValid() )
+    return 0;
+  else
+    return 6;
+}
+
+QVariant DictServersModel::headerData( int section, Qt::Orientation /*orientation*/, int role ) const
 {
   if ( role == Qt::DisplayRole )
     switch( section )
@@ -502,6 +816,12 @@ QVariant WebSitesModel::headerData( int section, Qt::Orientation /*orientation*/
         return tr( "Name" );
       case 2:
         return tr( "Address" );
+      case 3:
+        return tr( "Databases" );
+      case 4:
+        return tr( "Strategies" );
+      case 5:
+        return tr( "Icon" );
       default:
         return QVariant();
     }
@@ -509,9 +829,9 @@ QVariant WebSitesModel::headerData( int section, Qt::Orientation /*orientation*/
   return QVariant();
 }
 
-QVariant WebSitesModel::data( QModelIndex const & index, int role ) const
+QVariant DictServersModel::data( QModelIndex const & index, int role ) const
 {
-  if ( (unsigned)index.row() >= webSites.size() )
+  if ( index.row() >= dictServers.size() )
     return QVariant();
 
   if ( role == Qt::DisplayRole || role == Qt::EditRole )
@@ -519,33 +839,42 @@ QVariant WebSitesModel::data( QModelIndex const & index, int role ) const
     switch( index.column() )
     {
       case 1:
-        return webSites[ index.row() ].name;
+        return dictServers[ index.row() ].name;
       case 2:
-        return webSites[ index.row() ].url;
+        return dictServers[ index.row() ].url;
+      case 3:
+        return dictServers[ index.row() ].databases;
+      case 4:
+        return dictServers[ index.row() ].strategies;
+      case 5:
+        return dictServers[ index.row() ].iconFilename;
       default:
         return QVariant();
     }
   }
 
+  if( role == Qt::ToolTipRole && index.column() == 3 )
+    return tr( "Comma-delimited list of databases\n(empty string or \"*\" matches all databases)" );
+
+  if( role == Qt::ToolTipRole && index.column() == 4 )
+    return tr( "Comma-delimited list of search strategies\n(empty string mean \"prefix\" strategy)" );
+
   if ( role == Qt::CheckStateRole && !index.column() )
-    return webSites[ index.row() ].enabled;
+    return dictServers[ index.row() ].enabled ? Qt::Checked : Qt::Unchecked;
 
   return QVariant();
 }
 
-bool WebSitesModel::setData( QModelIndex const & index, const QVariant & value,
+bool DictServersModel::setData( QModelIndex const & index, const QVariant & value,
                                int role )
 {
-  if ( (unsigned)index.row() >= webSites.size() )
+  if ( index.row() >= dictServers.size() )
     return false;
 
   if ( role == Qt::CheckStateRole && !index.column() )
   {
-    //printf( "type = %d\n", (int)value.type() );
-    //printf( "value = %d\n", (int)value.toInt() );
-
     // XXX it seems to be always passing Int( 2 ) as a value, so we just toggle
-    webSites[ index.row() ].enabled = !webSites[ index.row() ].enabled;
+    dictServers[ index.row() ].enabled = !dictServers[ index.row() ].enabled;
 
     dataChanged( index, index );
     return true;
@@ -555,11 +884,23 @@ bool WebSitesModel::setData( QModelIndex const & index, const QVariant & value,
     switch( index.column() )
     {
       case 1:
-        webSites[ index.row() ].name =  value.toString();
+        dictServers[ index.row() ].name =  value.toString();
         dataChanged( index, index );
         return true;
       case 2:
-        webSites[ index.row() ].url =  value.toString();
+        dictServers[ index.row() ].url =  value.toString();
+        dataChanged( index, index );
+        return true;
+      case 3:
+        dictServers[ index.row() ].databases =  value.toString();
+        dataChanged( index, index );
+        return true;
+      case 4:
+        dictServers[ index.row() ].strategies =  value.toString();
+        dataChanged( index, index );
+        return true;
+      case 5:
+        dictServers[ index.row() ].iconFilename =  value.toString();
         dataChanged( index, index );
         return true;
       default:
@@ -570,6 +911,202 @@ bool WebSitesModel::setData( QModelIndex const & index, const QVariant & value,
 }
 
 
+////////// ProgramsModel
+
+ProgramsModel::ProgramsModel( QWidget * parent,
+                              Config::Programs const & programs_ ):
+  QAbstractItemModel( parent ), programs( programs_ )
+{
+}
+
+void ProgramsModel::removeProgram( int index )
+{
+  beginRemoveRows( QModelIndex(), index, index );
+  programs.erase( programs.begin() + index );
+  endRemoveRows();
+}
+
+void ProgramsModel::addNewProgram()
+{
+  Config::Program p;
+
+  p.enabled = false;
+  p.type = Config::Program::Audio;
+
+  p.id = Dictionary::generateRandomDictionaryId();
+
+  beginInsertRows( QModelIndex(), programs.size(), programs.size() );
+  programs.push_back( p );
+  endInsertRows();
+}
+
+QModelIndex ProgramsModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
+{
+  return createIndex( row, column );
+}
+
+QModelIndex ProgramsModel::parent( QModelIndex const & /*parent*/ ) const
+{
+  return QModelIndex();
+}
+
+Qt::ItemFlags ProgramsModel::flags( QModelIndex const & index ) const
+{
+  Qt::ItemFlags result = QAbstractItemModel::flags( index );
+
+  if ( index.isValid() )
+  {
+    if ( !index.column() )
+      result |= Qt::ItemIsUserCheckable;
+    else
+      result |= Qt::ItemIsEditable;
+  }
+
+  return result;
+}
+
+int ProgramsModel::rowCount( QModelIndex const & parent ) const
+{
+  if ( parent.isValid() )
+    return 0;
+  else
+    return programs.size();
+}
+
+int ProgramsModel::columnCount( QModelIndex const & parent ) const
+{
+  if ( parent.isValid() )
+    return 0;
+  else
+    return 5;
+}
+
+QVariant ProgramsModel::headerData( int section, Qt::Orientation /*orientation*/, int role ) const
+{
+  if ( role == Qt::DisplayRole )
+    switch( section )
+    {
+      case 0:
+        return tr( "Enabled" );
+      case 1:
+        return tr( "Type" );
+      case 2:
+        return tr( "Name" );
+      case 3:
+        return tr( "Command Line" );
+      case 4:
+        return tr( "Icon" );
+      default:
+        return QVariant();
+    }
+
+  return QVariant();
+}
+
+QVariant ProgramsModel::data( QModelIndex const & index, int role ) const
+{
+  if ( index.row() >= programs.size() )
+    return QVariant();
+
+  if ( role == Qt::DisplayRole || role == Qt::EditRole )
+  {
+    switch( index.column() )
+    {
+      case 1:
+        if ( role == Qt::DisplayRole )
+          return ProgramTypeEditor::getNameForType( programs[ index.row() ].type );
+        else
+          return QVariant( ( int ) programs[ index.row() ].type );
+      case 2:
+        return programs[ index.row() ].name;
+      case 3:
+        return programs[ index.row() ].commandLine;
+      case 4:
+        return programs[ index.row() ].iconFilename;
+      default:
+        return QVariant();
+    }
+  }
+
+  if ( role == Qt::CheckStateRole && !index.column() )
+    return programs[ index.row() ].enabled ? Qt::Checked : Qt::Unchecked;
+
+  return QVariant();
+}
+
+bool ProgramsModel::setData( QModelIndex const & index, const QVariant & value,
+                             int role )
+{
+  if ( index.row() >= programs.size() )
+    return false;
+
+  if ( role == Qt::CheckStateRole && !index.column() )
+  {
+    programs[ index.row() ].enabled = !programs[ index.row() ].enabled;
+
+    dataChanged( index, index );
+    return true;
+  }
+
+  if ( role == Qt::DisplayRole || role == Qt::EditRole )
+    switch( index.column() )
+    {
+      case 1:
+        programs[ index.row() ].type = Config::Program::Type( value.toInt() );
+        dataChanged( index, index );
+        return true;
+      case 2:
+        programs[ index.row() ].name = value.toString();
+        dataChanged( index, index );
+        return true;
+      case 3:
+        programs[ index.row() ].commandLine = value.toString();
+        dataChanged( index, index );
+        return true;
+      case 4:
+        programs[ index.row() ].iconFilename = value.toString();
+        dataChanged( index, index );
+        return true;
+      default:
+        return false;
+    }
+
+  return false;
+}
+
+QString ProgramTypeEditor::getNameForType( int v )
+{
+  switch( v )
+  {
+    case Config::Program::Audio:
+      return tr( "Audio" );
+    case Config::Program::PlainText:
+      return tr( "Plain Text" );
+    case Config::Program::Html:
+      return tr( "Html" );
+    case Config::Program::PrefixMatch:
+      return tr( "Prefix Match" );
+    default:
+      return tr( "Unknown" );
+  }
+}
+
+ProgramTypeEditor::ProgramTypeEditor( QWidget * widget ): QComboBox( widget )
+{
+  for( int x = 0; x < Config::Program::MaxTypeValue; ++x )
+    addItem( getNameForType( x ) );
+}
+
+int ProgramTypeEditor::getType() const
+{
+  return currentIndex();
+}
+
+void ProgramTypeEditor::setType( int t )
+{
+  setCurrentIndex( t );
+}
+
 ////////// PathsModel
 
 PathsModel::PathsModel( QWidget * parent,
@@ -594,7 +1131,7 @@ void PathsModel::addNewPath( QString const & path )
 
 QModelIndex PathsModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
 {
-  return createIndex( row, column, 0 );
+  return createIndex( row, column );
 }
 
 QModelIndex PathsModel::parent( QModelIndex const & /*parent*/ ) const
@@ -646,14 +1183,14 @@ QVariant PathsModel::headerData( int section, Qt::Orientation /*orientation*/, i
 
 QVariant PathsModel::data( QModelIndex const & index, int role ) const
 {
-  if ( (unsigned)index.row() >= paths.size() )
+  if ( index.row() >= paths.size() )
     return QVariant();
 
   if ( ( role == Qt::DisplayRole || role == Qt::EditRole ) && !index.column() )
     return paths[ index.row() ].path;
 
   if ( role == Qt::CheckStateRole && index.column() == 1 )
-    return paths[ index.row() ].recursive;
+    return paths[ index.row() ].recursive ? Qt::Checked : Qt::Unchecked;
 
   return QVariant();
 }
@@ -661,7 +1198,7 @@ QVariant PathsModel::data( QModelIndex const & index, int role ) const
 bool PathsModel::setData( QModelIndex const & index, const QVariant & /*value*/,
                           int role )
 {
-  if ( (unsigned)index.row() >= paths.size() )
+  if ( index.row() >= paths.size() )
     return false;
 
   if ( role == Qt::CheckStateRole && index.column() == 1 )
@@ -700,7 +1237,7 @@ void SoundDirsModel::addNewSoundDir( QString const & path, QString const & name
 
 QModelIndex SoundDirsModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
 {
-  return createIndex( row, column, 0 );
+  return createIndex( row, column );
 }
 
 QModelIndex SoundDirsModel::parent( QModelIndex const & /*parent*/ ) const
@@ -712,7 +1249,7 @@ Qt::ItemFlags SoundDirsModel::flags( QModelIndex const & index ) const
 {
   Qt::ItemFlags result = QAbstractItemModel::flags( index );
 
-  if ( index.isValid() && index.column() < 2 )
+  if ( index.isValid() && index.column() < 3 )
     result |= Qt::ItemIsEditable;
 
   return result;
@@ -731,7 +1268,7 @@ int SoundDirsModel::columnCount( QModelIndex const & parent ) const
   if ( parent.isValid() )
     return 0;
   else
-    return 2;
+    return 3;
 }
 
 QVariant SoundDirsModel::headerData( int section, Qt::Orientation /*orientation*/, int role ) const
@@ -743,6 +1280,8 @@ QVariant SoundDirsModel::headerData( int section, Qt::Orientation /*orientation*
         return tr( "Path" );
       case 1:
         return tr( "Name" );
+      case 2:
+        return tr( "Icon" );
       default:
         return QVariant();
     }
@@ -752,7 +1291,7 @@ QVariant SoundDirsModel::headerData( int section, Qt::Orientation /*orientation*
 
 QVariant SoundDirsModel::data( QModelIndex const & index, int role ) const
 {
-  if ( (unsigned)index.row() >= soundDirs.size() )
+  if ( index.row() >= soundDirs.size() )
     return QVariant();
 
   if ( ( role == Qt::DisplayRole || role == Qt::EditRole ) && !index.column() )
@@ -761,21 +1300,26 @@ QVariant SoundDirsModel::data( QModelIndex const & index, int role ) const
   if ( ( role == Qt::DisplayRole || role == Qt::EditRole ) && index.column() == 1 )
     return soundDirs[ index.row() ].name;
 
+  if ( ( role == Qt::DisplayRole || role == Qt::EditRole ) && index.column() == 2 )
+    return soundDirs[ index.row() ].iconFilename;
+
   return QVariant();
 }
 
 bool SoundDirsModel::setData( QModelIndex const & index, const QVariant & value,
                               int role )
 {
-  if ( (unsigned)index.row() >= soundDirs.size() )
+  if ( index.row() >= soundDirs.size() )
     return false;
 
-  if ( ( role == Qt::DisplayRole || role == Qt::EditRole ) && index.column() < 2 )
+  if ( ( role == Qt::DisplayRole || role == Qt::EditRole ) && index.column() < 3 )
   {
     if ( !index.column() )
       soundDirs[ index.row() ].path = value.toString();
-    else
+    else if ( index.column() == 1 )
       soundDirs[ index.row() ].name = value.toString();
+    else
+      soundDirs[ index.row() ].iconFilename = value.toString();
 
     dataChanged( index, index );
     return true;
@@ -797,12 +1341,13 @@ HunspellDictsModel::HunspellDictsModel( QWidget * parent,
 void HunspellDictsModel::changePath( QString const & newPath )
 {
   dataFiles = HunspellMorpho::findDataFiles( newPath );
-  reset();
+  beginResetModel();
+  endResetModel();
 }
 
 QModelIndex HunspellDictsModel::index( int row, int column, QModelIndex const & /*parent*/ ) const
 {
-  return createIndex( row, column, 0 );
+  return createIndex( row, column );
 }
 
 QModelIndex HunspellDictsModel::parent( QModelIndex const & /*parent*/ ) const
@@ -868,10 +1413,10 @@ QVariant HunspellDictsModel::data( QModelIndex const & index, int role ) const
     for( unsigned x = enabledDictionaries.size(); x--; )
     {
       if ( enabledDictionaries[ x ] == dataFiles[ index.row() ].dictId )
-        return true;
+        return Qt::Checked;
     }
 
-    return false;
+    return Qt::Unchecked;
   }
 
   return QVariant();
@@ -911,4 +1456,3 @@ void Sources::on_rescan_clicked()
 {
   emit rescan();
 }
-
